# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import logging
logger = logging.getLogger(__name__)

from mvc.models.properties import (
    IntegerProperty, ReadOnlyMixin, FloatProperty, LabeledProperty
)

from pyxrd.generic.models import PyXRDLine, ChildModel
from pyxrd.calculations.statistics import Rpw, Rp, derive

class Statistics(ChildModel):
    # PROPERTIES:
    specimen = property(ChildModel.parent.fget, ChildModel.parent.fset)

    @IntegerProperty(default=0, label="Points", visible=False, mix_with=(ReadOnlyMixin,))
    def points(self):
        try:
            e_ex, e_ey, e_cx, e_cy = self.specimen.get_exclusion_xy() #@UnusedVariable
            return e_ex.size
        except: pass
        return 0

    Rp = FloatProperty(default=None, label="Rp", visible=True)
    Rwp = FloatProperty(default=None, label="Rwp", visible=True)
    Rpder = FloatProperty(default=None, label="Rpder", visible=True)

    residual_pattern = LabeledProperty(default=None, label="Residual pattern")
    der_exp_pattern = LabeledProperty(default=None, label="Derived experimental pattern")
    der_calc_pattern = LabeledProperty(default=None, label="Derived calculated pattern")
    der_residual_pattern = LabeledProperty(default=None, label="Derived residual pattern")

    # ------------------------------------------------------------
    #      Initialization and other internals
    # ------------------------------------------------------------
    def __init__(self, *args, **kwargs):
        super(Statistics, self).__init__(*args, **kwargs)
        self.observe_model(self.parent)

    # ------------------------------------------------------------
    #      Notifications of observable properties
    # ------------------------------------------------------------
    @ChildModel.observe("parent", assign=True, after=True)
    def on_parent_changed(self, model, prop_name, info):
        self.update_statistics()

    # ------------------------------------------------------------
    #      Methods & Functions
    # ------------------------------------------------------------
    def _get_experimental(self):
        if self.specimen is not None:
            x, y = self.specimen.experimental_pattern.get_xy_data()
            return x.copy(), y.copy()
        else:
            return None, None
    def _get_calculated(self):
        if self.specimen is not None:
            x, y = self.specimen.calculated_pattern.get_xy_data()
            return x.copy(), y.copy()
        else:
            return None, None

    def scale_factor_y(self, offset):
        return self.specimen.scale_factor_y(offset) if self.specimen else (1.0, offset)

    def update_statistics(self, derived=False):
        # Clear factors:
        self.Rp = 0
        self.Rwp = 0
        self.Rpder = 0

        # Setup lines if not yet done:
        if self.residual_pattern == None:
            self.residual_pattern = PyXRDLine(label="Residual", color="#000000", lw=0.5, parent=self)

        if self.der_exp_pattern == None:
            self.der_exp_pattern = PyXRDLine(label="Exp. 1st der.", color="#000000", lw=2, parent=self)

        if self.der_calc_pattern == None:
            self.der_calc_pattern = PyXRDLine(label="Calc. 1st der.", color="#AA0000", lw=2, parent=self)

        if self.der_residual_pattern == None:
            self.der_residual_pattern = PyXRDLine(label="1st der. residual", color="#AA00AA", lw=1, parent=self)

        # Get data:
        exp_x, exp_y = self._get_experimental()
        cal_x, cal_y = self._get_calculated()
        der_exp_y, der_cal_y = None, None
        del cal_x # don't need this, is the same as exp_x

        # Try to get statistics, if it fails, just clear and inform the user
        try:
            if cal_y is not None and exp_y is not None and cal_y.size > 0 and exp_y.size > 0:
                # Get the selector for areas to consider in the statistics:
                selector = self.specimen.get_exclusion_selector()

                if derived:
                    # Calculate and set first derivate patterns:
                    der_exp_y, der_cal_y = derive(exp_y), derive(cal_y)
                    self.der_exp_pattern.set_data(exp_x, der_exp_y)
                    self.der_calc_pattern.set_data(exp_x, der_cal_y)

                # Calculate and set residual pattern:
                self.residual_pattern.set_data(exp_x, exp_y - cal_y)
                if derived:
                    self.der_residual_pattern.set_data(exp_x, der_exp_y - der_cal_y)

                # Calculate 'included' R values:
                self.Rp = Rp(exp_y[selector], cal_y[selector])
                self.Rwp = Rpw(exp_y[selector], cal_y[selector])
                if derived:
                    self.Rpder = Rp(der_exp_y[selector], der_cal_y[selector])
            else:
                self.residual_pattern.clear()
                self.der_exp_pattern.clear()
                self.der_calc_pattern.clear()
        except:
            self.residual_pattern.clear()
            self.der_exp_pattern.clear()
            self.der_calc_pattern.clear()
            logger.error("Error occurred when trying to calculate statistics, aborting calculation!")
            raise

    pass # end of class



#!/usr/bin/env python2

""" Git Versioning Script

Will transform stdin to expand some keywords with git version/author/date information.

Specify --clean to remove this information before commit.

Setup:

1. Copy versioning.py into your git repository

2. Run:

 git config filter.versioning.smudge 'python versioning.py'
 git config filter.versioning.clean  'python versioning.py --clean'
 echo 'version.py filter=versioning' >> .gitattributes
 git add versioning.py


3. add a version.py file with this contents:

 __version__ = ""

"""

import sys
import subprocess
import re


def main():
    clean = False
    if len(sys.argv) > 1:
        if sys.argv[1] == '--clean':
            clean = True

    # initialise empty here. Otherwise: forkbomb through the git calls.
    subst_list = {
        "version": "",
    }

    for line in sys.stdin:
        if not clean:
            subst_list = {
                # '--dirty' could be added to the following, too, but is not supported everywhere
                "version": subprocess.check_output(['git', 'describe', '--always']),
            }
            for k, v in subst_list.iteritems():
                v = re.sub(r'[\n\r\t"\']', "", v)
                rexp = "__%s__\s*=[\s'\"]+" % k
                line = re.sub(rexp, "__%s__ = \"%s\"\n" % (k, v), line)
            sys.stdout.write(line)
        else:
            for k in subst_list:
                rexp = "__%s__\s*=.*" % k
                line = re.sub(rexp, "__%s__ = \"\"" % k, line)
            sys.stdout.write(line)


if __name__ == "__main__":
    main()



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from pkg_resources import resource_filename  # @UnresolvedImport

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

import numpy as np

from pyxrd.generic.views import BaseView
from pyxrd.generic.views.validators import FloatEntryValidator

class EditInSituMixtureView(BaseView):   
    builder = resource_filename(__name__, "glade/edit_mixture.glade")
    top = "edit_mixture"

    base_width = 4
    base_height = 5

    matrix_widget = "tbl_matrix"
    wrapper_widget = "tbl_wrapper"
    widget_format = "mixture_%s"

    def __init__(self, *args, **kwargs):
        BaseView.__init__(self, *args, **kwargs)

        self.parent.set_title("Edit In Situ Mixtures")
        self.matrix = self[self.matrix_widget]
        self.wrapper = self[self.wrapper_widget]

        self.labels = [ self["lbl_scales"], self["lbl_fractions"], self["lbl_phases"], self["lbl_bgshifts"], self["lbl_specimens"] ]

        self["scolled_window"].set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        self.matrix.connect("size-request", self.on_size_requested)

        self.reset_view()

    def reset_view(self):
        def remove(item):
            if not item in self.labels: self.matrix.remove(item)
        self.matrix.foreach(remove)
        self.matrix.resize(self.base_height, self.base_width)

        self.phase_inputs = []
        self.fraction_inputs = []
        self.specimen_combos = []
        self.scale_inputs = []
        self.bgs_inputs = []
        self.phase_combos = np.empty(shape=(0, 0), dtype=np.object_) # 2D list
        self.behav_combos = np.empty(shape=(0, 0), dtype=np.object_) # 2D list

        self.on_size_requested()

    def on_size_requested(self, *args):
        sr = self.matrix.size_request()
        self[self.top].set_size_request(sr[0] + 100, -1)

    def set_edit_view(self, view):
        if self._on_sr_id is not None and self.child_view is not None:
            self.child_view.disconnect(self._on_sr_id)
        self.edit_view = view
        self.child_view = view.get_top_widget()
        self._add_child_view(self.child_view, self[self.edit_view_container])
        if isinstance(self[self.edit_view_container], Gtk.ScrolledWindow):
            sr = self.child_view.get_size_request()
            self[self.edit_view_container].set_size_request(sr[0], -1)

    def update_all(self, fractions, scales, bgs):
        for i, fraction in enumerate(fractions):
            if not i >= len(self.fraction_inputs):
                self.fraction_inputs[i].set_text(str(fraction))
        for i, scale in enumerate(scales):
            if not i >= len(self.scale_inputs):
                self.scale_inputs[i].set_text(str(scale))
        for i, bgs in enumerate(bgs):
            if not i >= len(self.bgs_inputs):
                self.bgs_inputs[i].set_text(str(bgs))

    def add_phase_slot(self, 
           phase_store, behav_store, 
           del_phase_callback, label_callback, fraction_callback, 
           phase_visible_callback, phase_combo_callback,
           behav_visible_callback, behav_combo_callback, 
           label, fraction,
           phases, behavs):
        r, c = self.matrix.get_property('n_rows'), self.matrix.get_property('n_columns')
        self.matrix.resize(r + 1, c)

        del_icon = Gtk.Image()
        del_icon.set_from_stock ("192-circle-remove", Gtk.IconSize.SMALL_TOOLBAR)
        new_phase_del_btn = Gtk.Button()
        new_phase_del_btn.set_image(del_icon)
        rid = new_phase_del_btn.connect("clicked", del_phase_callback)
        setattr(new_phase_del_btn, "deleventid", rid)
        self.matrix.attach(new_phase_del_btn, 0, 1, r, r + 1, Gtk.AttachOptions.FILL, 0)

        new_phase_input = self._get_new_input(label, callback=label_callback)
        self.phase_inputs.append(new_phase_input)
        self.matrix.attach(new_phase_input, 1, 2, r, r + 1, Gtk.AttachOptions.EXPAND | Gtk.AttachOptions.FILL, 0)

        new_fraction_input = self._get_new_input(str(fraction), callback=fraction_callback)
        FloatEntryValidator(new_fraction_input)
        self.fraction_inputs.append(new_fraction_input)
        self.matrix.attach(new_fraction_input, 2, 3, r, r + 1, Gtk.AttachOptions.EXPAND | Gtk.AttachOptions.FILL, 0)

        self.phase_combos.resize((c - self.base_width, r + 1 - self.base_height))
        self.behav_combos.resize((c - self.base_width, r + 1 - self.base_height))
        for col in range(c - self.base_width):
            mcol, mrow = r - self.base_height, col
            self._add_new_phase_behav_combo(
                mrow, mcol,
                phase_store, phase_store.c_name, phases[mrow, mcol], phase_visible_callback, phase_combo_callback,
                behav_store, behav_store.c_name, behavs[mrow, mcol], behav_visible_callback, behav_combo_callback
            )

        self.wrapper.show_all()

    def add_specimen_slot(self,
            phase_store, behav_store, specimen_store, 
            del_specimen_callback, scale_callback, bgs_callback, specimen_callback, 
            phase_visible_callback, phase_combo_callback,
            behav_visible_callback, behav_combo_callback, 
            scale, bgs, specimen, 
            phases, behavs):
        r, c = self.matrix.get_property('n_rows'), self.matrix.get_property('n_columns')
        self.matrix.resize(r, c + 1)

        del_icon = Gtk.Image()
        del_icon.set_from_stock("192-circle-remove", Gtk.IconSize.SMALL_TOOLBAR)
        new_specimen_del_btn = Gtk.Button()
        new_specimen_del_btn.set_image(del_icon)
        rid = new_specimen_del_btn.connect("clicked", del_specimen_callback)
        setattr(new_specimen_del_btn, "deleventid", rid)
        self.matrix.attach(new_specimen_del_btn, c, c + 1, 0, 1, Gtk.AttachOptions.EXPAND | Gtk.AttachOptions.FILL, 0)

        new_specimen_combo = self._get_new_combo(specimen_store, specimen_store.c_name, specimen, None, specimen_callback)
        self.specimen_combos.append(new_specimen_combo)
        self.matrix.attach(new_specimen_combo, c, c + 1, 1, 2, Gtk.AttachOptions.EXPAND | Gtk.AttachOptions.FILL, 0)

        new_bgs_input = self._get_new_input(str(bgs), callback=bgs_callback)
        FloatEntryValidator(new_bgs_input)
        self.bgs_inputs.append(new_bgs_input)
        self.matrix.attach(new_bgs_input, c, c + 1, 2, 3, Gtk.AttachOptions.EXPAND | Gtk.AttachOptions.FILL, 0)

        new_scale_input = self._get_new_input(str(scale), callback=scale_callback)
        FloatEntryValidator(new_scale_input)
        self.scale_inputs.append(new_scale_input)
        self.matrix.attach(new_scale_input, c, c + 1, 3, 4, Gtk.AttachOptions.EXPAND | Gtk.AttachOptions.FILL, 0)

        self.phase_combos.resize((c + 1 - self.base_width, r - self.base_height))
        self.behav_combos.resize((c + 1 - self.base_width, r - self.base_height))
        for row in range(r - self.base_height):
            mcol, mrow = row, c - self.base_width
            self._add_new_phase_behav_combo(
                mrow, mcol,
                phase_store, phase_store.c_name, phases[mrow, mcol], phase_visible_callback, phase_combo_callback,
                behav_store, behav_store.c_name, behavs[mrow, mcol], behav_visible_callback, behav_combo_callback
            )
            
        self.wrapper.show_all()

    def _get_new_input(self, text="", width=7, callback=None):
        """
            Creates a new text input box.
        """
        new_input = Gtk.Entry()
        new_input.set_text(text)
        new_input.set_alignment(0.0)
        new_input.set_width_chars(width)
        if callback is not None: new_input.connect("changed", callback)
        return new_input

    def _add_new_phase_behav_combo(self, 
            r, c,
            phase_model, phase_text_column, default_phase, phase_visible_callback, phase_callback,
            behav_model, behav_text_column, default_behav, behav_visible_callback, behav_callback):
        """
            Creates new 'phase' and 'behaviour' combo boxes, and adds them to the table at
            the given row and column indices.
        """
        new_phase_combo = self._get_new_combo(phase_model, phase_text_column, default_phase, phase_visible_callback, phase_callback, r, c)
        self.phase_combos[r, c] = new_phase_combo
        
        new_behav_combo = self._get_new_combo(behav_model, behav_text_column, default_behav, behav_visible_callback, behav_callback, r, c)
        self.behav_combos[r, c] = new_behav_combo

        box = Gtk.HBox(spacing = 5)
        box.pack_start(new_phase_combo, True, True, padding = 2)
        box.pack_end(new_behav_combo, True, True, padding = 2)
       
        self.matrix.attach(box, self.base_width + r, self.base_width + r + 1, self.base_height + c, self.base_height + c + 1, Gtk.AttachOptions.EXPAND | Gtk.AttachOptions.FILL, 0)

    def _get_new_combo(self, model, text_column, default, visible_callback, callback, *args):
        """
            Creates a new combo box with the given model as ListStore, setting
            the given column as text column, the given default value set as 
            active row, and connecting the given callback with 'changed' signal.
        """
        combobox = Gtk.ComboBox(model)
        combobox.set_size_request(75, -1)
        cell = Gtk.CellRendererText()
        combobox.pack_start(cell, True)

        def combo_func(cmb, cell, model, itr, args):
            text = model.get_value(itr, text_column)
            cell.set_property('text', text)
            if callable(visible_callback):
                cell.set_property('sensitive', visible_callback(model, itr, args))
        combobox.set_cell_data_func(cell, combo_func, args)       
            
        if default is not None:
            index = model.on_get_path(default)[0]
            combobox.set_active(index)
        combobox.connect("changed", callback, *args)
        return combobox

    pass # end of class



#!/usr/bin/python

# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import unittest

from pkg_resources import resource_filename # @UnresolvedImport
from test.setup import SKIP_REFINEMENT_TEST

from pyxrd.file_parsers.json_parser import JSONParser

__all__ = [
    'TestRefinement',
]

# Requires properly working:
#  - Phase
#  - Specimen
#  - Project
#  - Mixture

class TestRefinement(unittest.TestCase):

    atom_type = None

    def setUp(self):
        self.project = JSONParser.parse(resource_filename("test.test_mixture", "test refinement.pyxrd"))
        self.mixture = self.project.mixtures[0]

    def tearDown(self):
        del self.mixture
        del self.project

    def test_not_none(self):
        self.assertIsNotNone(self.mixture)

    def test_data_object(self):
        self.assertIsNotNone(self.mixture.data_object)

    @unittest.skipIf(SKIP_REFINEMENT_TEST, "Skipping refinement test")
    def test_refine_methods(self):
        for index, method in enumerate(self.mixture.refinement.refine_methods):
            self.mixture.refinement.refine_method_index = index
            self.mixture.refinement.randomize()
            refiner = self.mixture.refinement.get_refiner()
            refiner.refine(stop=None)

    pass # end of class


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()


# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Author: Mathijs Dumon
# Based on code from Rick Spencer
# Copyright (c) 2013, Mathijs Dumon, Rick Spencer
# All rights reserved.
# Complete license can be found in the LICENSE file.

import logging
logger = logging.getLogger(__name__)

from threading import Thread, Event

class CancellableThread(Thread):
    """
        Class for use by ThreadedTaskBox.
    """
    def __init__(self, run_function, on_complete):
        Thread.__init__(self)
        self.setDaemon(True)
        self.run_function = run_function
        self.on_complete = on_complete
        # Internal flag, for checking wether the user stopped.
        self.__stop = Event()
        # Internal flag, for checking wether the user cancelled.
        self.__cancel = Event()

    def run(self):
        try:
            # Tell the function to run
            data = self.run_function(stop=self.__stop)
            # Return function results, if not cancelled
            if not self.__cancel.is_set():
                self.on_complete(data)
        except KeyboardInterrupt:
            self.cancel()
        except:
            logger.exception("Unhandled exception in CancellableThread run()")

    def stop(self):
        """
            Stops the thread, and calls the on_complete callback
        """
        self.__stop.set()

    def cancel(self):
        """
            Stops the thread, and does not call the on_complete callback.
        """
        self.__cancel.set()
        self.__stop.set()


    pass #end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from pyxrd.generic.models import ChildModel

from pyxrd.calculations.mixture import (
    calculate_mixture,
    calculate_and_optimize_mixture,
    get_residual,
    get_optimized_residual
)


class Optimizer(ChildModel):
    """
        A simple model that plugs onto the Mixture model. It provides
        the functionality related to optimizing the weight fractions, scales
        and background shifts and residual calculation for the phases.
    """
    parent_alias = "mixture"

    # ------------------------------------------------------------
    #      Methods & Functions
    # ------------------------------------------------------------
    def get_current_residual(self):
        """
            Gets the residual for the current mixture solution.
            Convenience function.
        """
        return self.get_residual()

    def get_optimized_residual(self, data_object=None):
        """
            Gets an optimized residual for the current mixture setup. If no
            data_object is passed it is retrieved from the mixture.
        """
        return get_optimized_residual(*self.get_data_object(data_object))[0]

    def get_residual(self, data_object=None):
        """
            Calculates the residual for the given solution in combination with
            the given optimization arguments. If no data_object is passed it is 
            retrieved from the mixture.
        """
        return get_residual(*self.get_data_object(data_object))[0]

    def calculate(self, data_object=None):
        """
            Calculates the total and phase intensities. If no data_object is
            passed it is retrieved from the mixture.
        """
        return calculate_mixture(*self.get_data_object(data_object))

    def optimize(self, data_object=None):
        """
            Optimizes the mixture fractions, scales and bg shifts and returns the
            optimized result. If no data_object is passed it is retrieved from
            the mixture.
        """
        try:
            return calculate_and_optimize_mixture(*self.get_data_object(data_object))
        except AssertionError:
            return None

    def get_data_object(self, data_object=None):
        return (data_object if data_object is not None else self.parent.data_object,)

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import csv
import logging
logger = logging.getLogger(__name__)

from .ascii_parser import ASCIIParser

class CSVBaseParser(ASCIIParser):
    """
        CSV parser base functionality
    """

    default_fmt_params = {
        "delimiter": ',',
        "doublequote": True,
        "escapechar": None,
        "quotechar": "\"",
        "quoting": csv.QUOTE_MINIMAL,
        "skipinitialspace": True,
        "strict": False
    }

    @classmethod
    def parse_raw_line(cls, line, conv, **fmt_params):
        """ Parses a single raw line (read as a string) """
        fmt_params = dict(cls.default_fmt_params, **fmt_params)
        for row in csv.reader([line, ], **fmt_params):
            return list(map(conv, row))
            break # stop after first 'line' (there should only be one anyway)

    @classmethod
    def sniff(cls, f, **fmt_params):
        """ CSV Dialect guessing - f needs to be a file object! """
        # Guess dialect, and then:
        # - update default class dialect parameters with the sniffed dialect
        # - override the obtained dialect with any user-passed parameters
        dialect = None
        has_header = True
        file_start = 0
        if f is not None:
            f.seek(file_start)
            while True:
                file_start = f.tell()
                line = f.readline()
                if not line: break
                if not line.strip().startswith("#"): 
                    break # first non-comment line, assume no more comments after this
            try:
                sniffer = csv.Sniffer()
                f.seek(file_start)
                has_header = sniffer.has_header(f.read(1024))
                f.seek(file_start)
                f.readline() # skip (potential) header as these are sometimes formatted differently
                dialect = sniffer.sniff(f.read(1024))
                f.seek(file_start)
            except:
                logger.warning("Errors encountered while sniffing CSV dialect!")
                pass # ignore failures
        default_fmt_params = dict(cls.default_fmt_params, **{
            param: getattr(dialect, param) for param in list(cls.default_fmt_params.keys()) if hasattr(dialect, param)
        })
        return dict(default_fmt_params, **fmt_params), has_header, file_start

    pass # end of class



#!/usr/bin/python

# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import unittest

from test.utils import create_object_attribute_test, mock_settings

from pyxrd.project.models import Project

__all__ = [
    'TestProject',
]

class TestProject(unittest.TestCase):

    project = None

    def setUp(self):
        mock_settings();
        self.project = Project(name="Test Project")

    def tearDown(self):
        del self.project

    def test_not_none(self):
        self.assertIsNotNone(self.project)

    test_name = create_object_attribute_test("project", "name", "Test Name")
    test_date = create_object_attribute_test("project", "date", "19/09/1987")
    test_description = create_object_attribute_test("project", "description", "Test Description")
    test_author = create_object_attribute_test("project", "author", "Test Author")
    test_layout_mode = create_object_attribute_test("project", "layout_mode", "FULL")
    test_display_marker_align = create_object_attribute_test("project", "display_marker_align", "right")
    test_display_marker_color = create_object_attribute_test("project", "display_marker_color", "#FF00FF")
    test_display_marker_base = create_object_attribute_test("project", "display_marker_base", 2)
    test_display_marker_top = create_object_attribute_test("project", "display_marker_top", 1)
    test_display_marker_top_offset = create_object_attribute_test("project", "display_marker_top_offset", 0.5)
    test_display_marker_angle = create_object_attribute_test("project", "display_marker_angle", 45.6)
    test_display_marker_style = create_object_attribute_test("project", "display_marker_style", "dashed")
    test_display_calc_color = create_object_attribute_test("project", "display_calc_color", "#FF0099")
    test_display_exp_color = create_object_attribute_test("project", "display_exp_color", "#9900FF")
    test_display_calc_lw = create_object_attribute_test("project", "display_calc_lw", 5)
    test_display_exp_lw = create_object_attribute_test("project", "display_exp_lw", 1)
    test_display_plot_offset = create_object_attribute_test("project", "display_plot_offset", 1.5)
    test_display_group_by = create_object_attribute_test("project", "display_group_by", 3)
    test_display_label_pos = create_object_attribute_test("project", "display_label_pos", 0.75)
    test_axes_xscale = create_object_attribute_test("project", "axes_xscale", 1)
    test_axes_xmin = create_object_attribute_test("project", "axes_xmin", 15)
    test_axes_xmax = create_object_attribute_test("project", "axes_xmax", 52)
    test_axes_xstretch = create_object_attribute_test("project", "axes_xstretch", True)
    test_axes_yscale = create_object_attribute_test("project", "axes_yscale", 1)
    test_axes_yvisible = create_object_attribute_test("project", "axes_yvisible", True)

    # TODO
    #  - addition of phases, specimens & atom_types
    #  - loading of phases, specimens & atom_types
    #  - testing inherit properties (markers)
    #  - testing initialization deprecated keywords etc.

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import logging
logger = logging.getLogger(__name__)

import time
from contextlib import contextmanager

from mvc.models.properties import (
    FloatProperty, BoolProperty, StringProperty, ListProperty,
    IntegerProperty, StringChoiceProperty, IntegerChoiceProperty, SignalMixin
)

from mvc.observers import ListObserver

from pyxrd.__version import __version__

from pyxrd.data import settings

from pyxrd.generic.models import DataModel
from pyxrd.generic.models.event_context_manager import EventContextManager
from pyxrd.generic.io import storables, Storable, get_case_insensitive_glob
from pyxrd.generic.utils import not_none

from pyxrd.atoms.models import AtomType
from pyxrd.phases.models import Phase
from pyxrd.specimen.models import Specimen
from pyxrd.mixture.models.mixture import Mixture
#from pyxrd.mixture.models.insitu_behaviours import InSituBehaviour

@storables.register()
class Project(DataModel, Storable):
    """
    This is the top-level object that servers the purpose of combining the
    different objects (most notably :class:`~.atoms.models.AtomType`'s,
    :class:`~.phases.models.Phase`'s, :class:`~.specimen.models.Specimen`'s and
    :class:`~.mixture.models.Mixture`'s).
    
    It also provides a large number of display-related 'default' properties 
    (e.g. for patterns and their markers, axes etc.). For more details: see the
    property descriptions.
    
    Example usage:
    
    .. code-block:: python
    
        >>> from pyxrd.project.models import Project
        >>> from pyxrd.generic.io.xrd_parsers import XRDParser
        >>> from pyxrd.specimen.models import Specimen
        >>> project = Project(name="New Project", author="Mr. X", layout_mode="FULL", axes_dspacing=True)
        >>> for specimen in Specimen.from_experimental_data("/path/to/xrd_data_file.rd", parent=project):
        ...   project.specimens.append(specimen)
        ...
        
    """

    # MODEL INTEL:
    class Meta(DataModel.Meta):
        store_id = "Project"
        file_filters = [
            ("PyXRD Project files", get_case_insensitive_glob("*.pyxrd", "*.zpd")),
        ]
        import_filters = [
            ("Sybilla XML files", get_case_insensitive_glob("*.xml")),
        ]

    # PROPERTIES:

    filename = None

    #: The project name
    name = StringProperty(
        default="", text="Name",
        visible=True, persistent=True
    )

    #: The project data (string)
    date = StringProperty(
        default="", text="Date",
        visible=True, persistent=True
    )

    #: The project description
    description = StringProperty(
        default=None, text="Description",
        visible=True, persistent=True, widget_type="text_view",
    )

    #: The project author
    author = StringProperty(
        default="", text="Author",
        visible=True, persistent=True
    )

    #: Flag indicating whether this project has been changed since it was last saved.
    needs_saving = BoolProperty(
        default=True, visible=False, persistent=False
    )

    #: The layout mode this project should be displayed in
    layout_mode = StringChoiceProperty(
        default=settings.DEFAULT_LAYOUT, text="Layout mode",
        visible=True, persistent=True, choices=settings.DEFAULT_LAYOUTS,
        mix_with=(SignalMixin,), signal_name="visuals_changed"
    )

    #: The manual lower limit for the X-axis
    axes_xmin = FloatProperty(
        default=settings.AXES_MANUAL_XMIN, text="min. [°2T]",
        visible=True, persistent=True, minimum=0.0, widget_type="spin",
        mix_with=(SignalMixin,), signal_name="visuals_changed"
    )

    #: The manual upper limit for the X-axis
    axes_xmax = FloatProperty(
        default=settings.AXES_MANUAL_XMAX, text="max. [°2T]",
        visible=True, persistent=True, minimum=0.0, widget_type="spin",
        mix_with=(SignalMixin,), signal_name="visuals_changed"
    )

    #: Whether or not to stretch the X-axis over the entire available display
    axes_xstretch = BoolProperty(
        default=settings.AXES_XSTRETCH, text="Stetch x-axis to fit window",
        visible=True, persistent=True,
        mix_with=(SignalMixin,), signal_name="visuals_changed"
    )

    #: Flag toggling between d-spacing (when True) or 2-Theta axes (when False)
    axes_dspacing = BoolProperty(
        default=settings.AXES_DSPACING, text="Show d-spacing in x-axis",
        visible=True, persistent=True,
        mix_with=(SignalMixin,), signal_name="visuals_changed"
    )

    #: Whether or not the y-axis should be shown
    axes_yvisible = BoolProperty(
        default=settings.AXES_YVISIBLE, text="Y-axis visible",
        visible=True, persistent=True,
        mix_with=(SignalMixin,), signal_name="visuals_changed"
    )

    #: The manual lower limit for the Y-axis (in counts)
    axes_ymin = FloatProperty(
        default=settings.AXES_MANUAL_YMIN, text="min. [counts]",
        visible=True, persistent=True, minimum=0.0, widget_type="spin",
        mix_with=(SignalMixin,), signal_name="visuals_changed"
    )

    #: The manual upper limit for the Y-axis (in counts)
    axes_ymax = FloatProperty(
        default=settings.AXES_MANUAL_YMAX, text="max. [counts]",
        visible=True, persistent=True, minimum=0.0, widget_type="spin",
        mix_with=(SignalMixin,), signal_name="visuals_changed"
    )

    #: What type of y-axis to use: raw counts, single or multi-normalized units
    axes_ynormalize = IntegerChoiceProperty(
        default=settings.AXES_YNORMALIZE, text="Y scaling",
        visible=True, persistent=True, choices=settings.AXES_YNORMALIZERS,
        mix_with=(SignalMixin,), signal_name="visuals_changed"
    )

    #: Whether to use automatic or manual Y limits
    axes_ylimit = IntegerChoiceProperty(
        default=settings.AXES_YLIMIT, text="Y limit",
        visible=True, persistent=True, choices=settings.AXES_YLIMITS,
        mix_with=(SignalMixin,), signal_name="visuals_changed"
    )

    #: The offset between patterns as a fraction of the maximum intensity
    display_plot_offset = FloatProperty(
        default=settings.PLOT_OFFSET, text="Pattern offset",
        visible=True, persistent=True, minimum=0.0, widget_type="float_entry",
        mix_with=(SignalMixin,), signal_name="visuals_changed"
    )

    #: The number of patterns to group ( = having no offset)
    display_group_by = IntegerProperty(
        default=settings.PATTERN_GROUP_BY, text="Group patterns by",
        visible=True, persistent=True, minimum=1, widget_type="spin",
        mix_with=(SignalMixin,), signal_name="visuals_changed"
    )

    #: The relative position (from the pattern offset) for pattern labels
    #: as a fraction of the patterns intensity
    display_label_pos = FloatProperty(
        default=settings.LABEL_POSITION, text="Default label position",
        visible=True, persistent=True, widget_type="float_entry",
        mix_with=(SignalMixin,), signal_name="visuals_changed"
    )

    #: What type of scale to use for X-axis, automatic or manual
    axes_xlimit = IntegerChoiceProperty(
        default=settings.AXES_XLIMIT, text="X limit",
        visible=True, persistent=True,
        choices=settings.AXES_XLIMITS,
        mix_with=(SignalMixin,), signal_name="visuals_changed"
    )

    #: The default angle at which marker labels are displayed
    display_marker_angle = FloatProperty(
        default=settings.MARKER_ANGLE, text="Angle",
        visible=True, persistent=True, widget_type="float_entry",
        mix_with=(SignalMixin,), signal_name="visuals_changed"
    )

    #: The default offset for marker labels
    display_marker_top_offset = FloatProperty(
        default=settings.MARKER_TOP_OFFSET, text="Offset from base",
        visible=True, persistent=True, widget_type="float_entry",
        mix_with=(SignalMixin,), signal_name="visuals_changed"
    )

    #: The default marker label alignment (one of settings.MARKER_ALIGNS)
    display_marker_align = StringChoiceProperty(
        default=settings.MARKER_ALIGN, text="Label alignment",
        visible=True, persistent=True,
        choices=settings.MARKER_ALIGNS,
        mix_with=(SignalMixin,), signal_name="visuals_changed"
    )

    #: The default marker label base (one of settings.MARKER_BASES)
    display_marker_base = IntegerChoiceProperty(
        default=settings.MARKER_BASE, text="Base connection",
        visible=True, persistent=True,
        choices=settings.MARKER_BASES,
        mix_with=(SignalMixin,), signal_name="visuals_changed"
    )

    #: The default marker label top (one of settings.MARKER_TOPS)
    display_marker_top = IntegerChoiceProperty(
        default=settings.MARKER_TOP, text="Top connection",
        visible=True, persistent=True,
        choices=settings.MARKER_TOPS,
        mix_with=(SignalMixin,), signal_name="visuals_changed"
    )

    #: The default marker style (one of settings.MARKER_STYLES)
    display_marker_style = StringChoiceProperty(
        default=settings.MARKER_STYLE, text="Line style",
        visible=True, persistent=True,
        choices=settings.MARKER_STYLES,
        mix_with=(SignalMixin,), signal_name="visuals_changed"
    )

    #: The default marker color
    display_marker_color = StringProperty(
        default=settings.MARKER_COLOR, text="Color",
        visible=True, persistent=True, widget_type="color",
        mix_with=(SignalMixin,), signal_name="visuals_changed"
    )

    #: The default calculated profile color
    display_calc_color = StringProperty(
        default=settings.CALCULATED_COLOR, text="Calculated color",
        visible=True, persistent=True, widget_type="color",
        mix_with=(SignalMixin,), signal_name="visuals_changed"
    )

    #: The default experimental profile color
    display_exp_color = StringProperty(
        default=settings.EXPERIMENTAL_COLOR, text="Experimental color",
        visible=True, persistent=True, widget_type="color",
        mix_with=(SignalMixin,), signal_name="visuals_changed"
    )

    #: The default calculated profile line width
    display_calc_lw = IntegerProperty(
        default=settings.CALCULATED_LINEWIDTH, text="Calculated line width",
        visible=True, persistent=True, widget_type="spin",
        mix_with=(SignalMixin,), signal_name="visuals_changed"
    )

    #: The default experimental profile line width
    display_exp_lw = IntegerProperty(
        default=settings.EXPERIMENTAL_LINEWIDTH, text="Experimental line width",
        visible=True, persistent=True, widget_type="spin",
        mix_with=(SignalMixin,), signal_name="visuals_changed"
    )

    #: The default calculated profile line style
    display_calc_ls = StringChoiceProperty(
        default=settings.CALCULATED_LINESTYLE, text="Calculated line style",
        visible=True, persistent=True,
        choices=settings.PATTERN_LINE_STYLES,
        mix_with=(SignalMixin,), signal_name="visuals_changed"
    )

    #: The default experimental profile line style
    display_exp_ls = StringChoiceProperty(
        default=settings.EXPERIMENTAL_LINESTYLE, text="Experimental line style",
        visible=True, persistent=True,
        choices=settings.PATTERN_LINE_STYLES,
        mix_with=(SignalMixin,), signal_name="visuals_changed"
    )

    #: The default calculated profile line style
    display_calc_marker = StringChoiceProperty(
        default=settings.CALCULATED_MARKER, text="Calculated line marker",
        visible=True, persistent=True,
        choices=settings.PATTERN_MARKERS,
        mix_with=(SignalMixin,), signal_name="visuals_changed"
    )

    #: The default calculated profile line style
    display_exp_marker = StringChoiceProperty(
        default=settings.EXPERIMENTAL_MARKER, text="Experimental line marker",
        visible=True, persistent=True,
        choices=settings.PATTERN_MARKERS,
        mix_with=(SignalMixin,), signal_name="visuals_changed"
    )

    #: The list of specimens
    specimens = ListProperty(
        default=[], text="Specimens", data_type=Specimen,
        visible=True, persistent=True,
    )

    #: The list of phases
    phases = ListProperty(
        default=[], text="Phases", data_type=Phase,
        visible=False, persistent=True
    )

    #: The list of atom types
    atom_types = ListProperty(
        default=[], text="Atom types", data_type=AtomType,
        visible=False, persistent=True
    )

    #: The list of Behaviours
    #behaviours = ListProperty(
    #    default=[], text="Behaviours", data_type=InSituBehaviour,
    #    visible=False, persistent=True
    #)

    #: The list of mixtures
    mixtures = ListProperty(
        default=[], text="Mixture", data_type=Mixture,
        visible=False, persistent=True
    )

    # ------------------------------------------------------------
    #      Initialization and other internals
    # ------------------------------------------------------------
    def __init__(self, *args, **kwargs):
        """
            Constructor takes any of its properties as a keyword argument 
            except for:
                - needs_saving
            
            In addition to the above, the constructor still supports the 
            following deprecated keywords, mapping to a current keyword:
                - goniometer: the project-level goniometer, is passed on to the
                  specimens
                - axes_xscale: deprecated alias for axes_xlimit
                - axes_yscale: deprecated alias for axes_ynormalize
                
            Any other arguments or keywords are passed to the base class.
        """
        my_kwargs = self.pop_kwargs(kwargs,
            "goniometer", "data_goniometer", "data_atom_types", "data_phases",
            "axes_yscale", "axes_xscale", "filename", "behaviours",
            *[prop.label for prop in Project.Meta.get_local_persistent_properties()]
        )
        super(Project, self).__init__(*args, **kwargs)
        kwargs = my_kwargs

        with self.data_changed.hold():
            with self.visuals_changed.hold():

                self.filename = self.get_kwarg(kwargs, self.filename, "filename")
                self.layout_mode = self.get_kwarg(kwargs, self.layout_mode, "layout_mode")

                self.display_marker_align = self.get_kwarg(kwargs, self.display_marker_align, "display_marker_align")
                self.display_marker_color = self.get_kwarg(kwargs, self.display_marker_color, "display_marker_color")
                self.display_marker_base = self.get_kwarg(kwargs, self.display_marker_base, "display_marker_base")
                self.display_marker_top = self.get_kwarg(kwargs, self.display_marker_top, "display_marker_top")
                self.display_marker_top_offset = self.get_kwarg(kwargs, self.display_marker_top_offset, "display_marker_top_offset")
                self.display_marker_angle = self.get_kwarg(kwargs, self.display_marker_angle, "display_marker_angle")
                self.display_marker_style = self.get_kwarg(kwargs, self.display_marker_style, "display_marker_style")

                self.display_calc_color = self.get_kwarg(kwargs, self.display_calc_color, "display_calc_color")
                self.display_exp_color = self.get_kwarg(kwargs, self.display_exp_color, "display_exp_color")
                self.display_calc_lw = self.get_kwarg(kwargs, self.display_calc_lw, "display_calc_lw")
                self.display_exp_lw = self.get_kwarg(kwargs, self.display_exp_lw, "display_exp_lw")
                self.display_calc_ls = self.get_kwarg(kwargs, self.display_calc_ls, "display_calc_ls")
                self.display_exp_ls = self.get_kwarg(kwargs, self.display_exp_ls, "display_exp_ls")
                self.display_calc_marker = self.get_kwarg(kwargs, self.display_calc_marker, "display_calc_marker")
                self.display_exp_marker = self.get_kwarg(kwargs, self.display_exp_marker, "display_exp_marker")
                self.display_plot_offset = self.get_kwarg(kwargs, self.display_plot_offset, "display_plot_offset")
                self.display_group_by = self.get_kwarg(kwargs, self.display_group_by, "display_group_by")
                self.display_label_pos = self.get_kwarg(kwargs, self.display_label_pos, "display_label_pos")

                self.axes_xlimit = self.get_kwarg(kwargs, self.axes_xlimit, "axes_xlimit", "axes_xscale")
                self.axes_xmin = self.get_kwarg(kwargs, self.axes_xmin, "axes_xmin")
                self.axes_xmax = self.get_kwarg(kwargs, self.axes_xmax, "axes_xmax")
                self.axes_xstretch = self.get_kwarg(kwargs, self.axes_xstretch, "axes_xstretch")
                self.axes_ylimit = self.get_kwarg(kwargs, self.axes_ylimit, "axes_ylimit")
                self.axes_ynormalize = self.get_kwarg(kwargs, self.axes_ynormalize, "axes_ynormalize", "axes_yscale")
                self.axes_yvisible = self.get_kwarg(kwargs, self.axes_yvisible, "axes_yvisible")
                self.axes_ymin = self.get_kwarg(kwargs, self.axes_ymin, "axes_ymin")
                self.axes_ymax = self.get_kwarg(kwargs, self.axes_ymax, "axes_ymax")

                goniometer = None
                goniometer_kwargs = self.get_kwarg(kwargs, None, "goniometer", "data_goniometer")
                if goniometer_kwargs:
                    goniometer = self.parse_init_arg(goniometer_kwargs, None, child=True)

                # Set up and observe atom types:
                self.atom_types = self.get_list(kwargs, [], "atom_types", "data_atom_types", parent=self)
                self._atom_types_observer = ListObserver(
                    self.on_atom_type_inserted,
                    self.on_atom_type_removed,
                    prop_name="atom_types",
                    model=self
                )

                # Resolve json references & observe phases
                self.phases = self.get_list(kwargs, [], "phases", "data_phases", parent=self)
                for phase in self.phases:
                    phase.resolve_json_references()
                    self.observe_model(phase)
                self._phases_observer = ListObserver(
                    self.on_phase_inserted,
                    self.on_phase_removed,
                    prop_name="phases",
                    model=self
                )

                # Set goniometer if required & observe specimens
                self.specimens = self.get_list(kwargs, [], "specimens", "data_specimens", parent=self)
                for specimen in self.specimens:
                    if goniometer: specimen.goniometer = goniometer
                    self.observe_model(specimen)
                self._specimens_observer = ListObserver(
                    self.on_specimen_inserted,
                    self.on_specimen_removed,
                    prop_name="specimens",
                    model=self
                )

                # Observe behaviours:
                #self.behaviours = self.get_list(kwargs, [], "behaviours", parent=self)
                #for behaviour in self.behaviours:
                #    self.observe_model(behaviour)
                #self._behaviours_observer = ListObserver(
                #    self.on_behaviour_inserted,
                #    self.on_behaviour_removed,
                #    prop_name="behaviours",
                #    model=self
                #)

                # Observe mixtures:
                self.mixtures = self.get_list(kwargs, [], "mixtures", "data_mixtures", parent=self)
                for mixture in self.mixtures:
                    self.observe_model(mixture)
                self._mixtures_observer = ListObserver(
                    self.on_mixture_inserted,
                    self.on_mixture_removed,
                    prop_name="mixtures",
                    model=self
                )

                self.name = str(self.get_kwarg(kwargs, "Project name", "name", "data_name"))
                self.date = str(self.get_kwarg(kwargs, time.strftime("%d/%m/%Y"), "date", "data_date"))
                self.description = str(self.get_kwarg(kwargs, "Project description", "description", "data_description"))
                self.author = str(self.get_kwarg(kwargs, "Project author", "author", "data_author"))

                load_default_data = self.get_kwarg(kwargs, True, "load_default_data")
                if load_default_data and self.layout_mode != 1 and \
                    len(self.atom_types) == 0: self.load_default_data()

                self.needs_saving = True
            pass # end with visuals_changed
        pass # end with data_changed

    def load_default_data(self):
        for atom_type in AtomType.get_from_csv(settings.DATA_REG.get_file_path("ATOM_SCAT_FACTORS")):
            self.atom_types.append(atom_type)

    # ------------------------------------------------------------
    #      Notifications of observable properties
    # ------------------------------------------------------------
    def on_phase_inserted(self, item):
        # Set parent on the new phase:
        if item.parent != self: item.parent = self
        item.resolve_json_references()

    def on_phase_removed(self, item):
        with self.data_changed.hold_and_emit():
            # Clear parent:
            item.parent = None
            # Clear links with other phases:
            if getattr(item, "based_on", None) is not None:
                item.based_on = None
            for phase in self.phases:
                if getattr(phase, "based_on", None) == item:
                    phase.based_on = None
            # Remove phase from mixtures:
            for mixture in self.mixtures:
                mixture.unset_phase(item)

    def on_atom_type_inserted(self, item, *data):
        if item.parent != self: item.parent = self
        # We do not observe AtomType's directly, if they change,
        # Atoms containing them will be notified, and that event should bubble
        # up to the project level.

    def on_atom_type_removed(self, item, *data):
        item.parent = None
        # We do not emit a signal for AtomType's, if it was part of
        # an Atom, the Atom will be notified, and the event should bubble
        # up to the project level

    def on_specimen_inserted(self, item):
        # Set parent and observe the new specimen (visuals changed signals):
        if item.parent != self: item.parent = self
        self.observe_model(item)

    def on_specimen_removed(self, item):
        with self.data_changed.hold_and_emit():
            # Clear parent & stop observing:
            item.parent = None
            self.relieve_model(item)
            # Remove specimen from mixtures:
            for mixture in self.mixtures:
                mixture.unset_specimen(item)

    def on_mixture_inserted(self, item):
        # Set parent and observe the new mixture:
        if item.parent != self: item.parent = self
        self.observe_model(item)

    def on_mixture_removed(self, item):
        with self.data_changed.hold_and_emit():
            # Clear parent & stop observing:
            item.parent = None
            self.relieve_model(item)

    def on_behaviour_inserted(self, item):
        # Set parent and observe the new mixture:
        if item.parent != self: item.parent = self
        self.observe_model(item)

    def on_behaviour_removed(self, item):
        with self.data_changed.hold_and_emit():
            # Clear parent & stop observing:
            item.parent = None
            self.relieve_model(item)

    @DataModel.observe("data_changed", signal=True)
    def notify_data_changed(self, model, prop_name, info):
        self.needs_saving = True
        if isinstance(model, Mixture):
            self.data_changed.emit()

    @DataModel.observe("visuals_changed", signal=True)
    def notify_visuals_changed(self, model, prop_name, info):
        self.needs_saving = True
        self.visuals_changed.emit() # propagate signal

    # ------------------------------------------------------------
    #      Input/Output stuff
    # ------------------------------------------------------------
    @classmethod
    def from_json(type, **kwargs): # @ReservedAssignment
        project = type(**kwargs)
        project.needs_saving = False # don't mark this when just loaded
        return project

    def to_json_multi_part(self):
        to_json = self.to_json()
        properties = to_json["properties"]

        for name in ("phases", "specimens", "atom_types", "mixtures"): #"behaviours"
            yield (name, properties.pop(name))
            properties[name] = "file://%s" % name

        yield ("content", to_json)
        yield ("version", __version__)

    @staticmethod
    def create_from_sybilla_xml(filename, **kwargs):
        from pyxrd.project.importing import create_project_from_sybilla_xml
        return create_project_from_sybilla_xml(filename, **kwargs)

    # ------------------------------------------------------------
    #      Draggable mix-in hook:
    # ------------------------------------------------------------
    def on_label_dragged(self, delta_y, button=1):
        if button == 1:
            self.display_label_pos += delta_y
        pass

    # ------------------------------------------------------------
    #      Methods & Functions
    # ------------------------------------------------------------
    def get_scale_factor(self, specimen=None):
        """
        Get the factor with which to scale raw data and the scaled offset
                
        :rtype: tuple containing the scale factor and the (scaled) offset
        """
        if self.axes_ynormalize == 0 or (self.axes_ynormalize == 1 and specimen is None):
            return (1.0 / (self.get_max_display_y() or 1.0), 1.0)
        elif self.axes_ynormalize == 1:
            return (1.0 / (specimen.get_max_display_y or 1.0), 1.0)
        elif self.axes_ynormalize == 2:
            return (1.0, self.get_max_display_y())
        else:
            raise ValueError("Wrong value for 'axes_ysnormalize' in %s: is `%d`; should be 0, 1 or 2" % (self, self.axes_ynormalize))

    def get_max_display_y(self):
        max_display_y = 0
        if self.parent is not None:
            for specimen in self.parent.current_specimens:
                max_display_y = max(specimen.max_display_y, max_display_y)
        return max_display_y

    @contextmanager
    def hold_child_signals(self):
        logger.info("Holding back all project child object signals")
        with self.hold_mixtures_needs_update():
            with self.hold_mixtures_data_changed():
                with self.hold_phases_data_changed():
                    with self.hold_specimens_data_changed():
                        with self.hold_atom_types_data_changed():
                            yield

    @contextmanager
    def hold_mixtures_needs_update(self):
        logger.info("Holding back all 'needs_update' signals from Mixtures")
        with EventContextManager(*[mixture.needs_update.hold() for mixture in self.mixtures]):
            yield

    @contextmanager
    def hold_mixtures_data_changed(self):
        logger.info("Holding back all 'data_changed' signals from Mixtures")
        with EventContextManager(*[mixture.data_changed.hold() for mixture in self.mixtures]):
            yield

    @contextmanager
    def hold_phases_data_changed(self):
        logger.info("Holding back all 'data_changed' signals from Phases")
        with EventContextManager(*[phase.data_changed.hold() for phase in self.phases]):
            yield

    @contextmanager
    def hold_atom_types_data_changed(self):
        logger.info("Holding back all 'data_changed' signals from AtomTypes")
        with EventContextManager(*[atom_type.data_changed.hold() for atom_type in self.atom_types]):
            yield

    @contextmanager
    def hold_specimens_data_changed(self):
        logger.info("Holding back all 'data_changed' signals from Specimens")
        with EventContextManager(*[specimen.data_changed.hold() for specimen in self.specimens]):
            yield

    def update_all_mixtures(self):
        """
        Forces all mixtures in this project to update. If they have auto
        optimization enabled, this will also optimize them. 
        """
        for mixture in self.mixtures:
            with self.data_changed.ignore():
                mixture.update()

    def get_mixtures_by_name(self, mixture_name):
        """
        Convenience method that returns all the mixtures who's name match the
        passed name as a list.
        """
        return [mixture for mixture in self.mixtures if (mixture.name == mixture_name)]

    # ------------------------------------------------------------
    #      Specimen list related
    # ------------------------------------------------------------
    def move_specimen_up(self, specimen):
        """
        Move the passed :class:`~pyxrd.specimen.models.Specimen` up one slot.
        Will raise and IndexError if the passed specimen is not in this project.
        """
        index = self.specimens.index(specimen)
        self.specimens.insert(min(index + 1, len(self.specimens)), self.specimens.pop(index))

    def move_specimen_down(self, specimen):
        """
        Move the passed :class:`~pyxrd.specimen.models.Specimen` down one slot
        Will raise and IndexError if the passed specimen is not in this project.
        """
        index = self.specimens.index(specimen)
        self.specimens.insert(max(index - 1, 0), self.specimens.pop(index))
        pass

    # ------------------------------------------------------------
    #      Phases list related
    # ------------------------------------------------------------
    def load_phases(self, filename, parser, insert_index=0):
        """
        Loads all :class:`~pyxrd.phase.models.Phase` objects from the file
        'filename'. An optional index can be given where the phases need to be
        inserted at.
        """
        # make sure we have no duplicate UUID's
        insert_index = not_none(insert_index, 0)
        type(Project).object_pool.change_all_uuids()
        for phase in parser.parse(filename):
            phase.parent = self
            self.phases.insert(insert_index, phase)
            insert_index += 1

    # ------------------------------------------------------------
    #      AtomType's list related
    # ------------------------------------------------------------
    def load_atom_types(self, filename, parser):
        """
        Loads all :class:`~pyxrd.atoms.models.AtomType` objects from the
        file specified by *filename*.
        """
        # make sure we have no duplicate UUID's
        type(Project).object_pool.change_all_uuids()
        for atom_type in parser.parse(filename):
            atom_type.parent = self
            self.atom_types.append(atom_type)

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk

import logging
logger = logging.getLogger(__name__)

from pyxrd.generic.controllers import DialogController

class AddMixtureController(DialogController):
    """ 
        Controller for the add mixture dialog
    """

    auto_adapt = False

    def __init__(self, model=None, view=None, parent=None, callback=None):
        super(AddMixtureController, self).__init__(
            model=model, view=view, parent=parent)
        self.callback = callback

    # ------------------------------------------------------------
    #      GTK Signal handlers
    # ------------------------------------------------------------
    def on_btn_ok_clicked(self, event):
        self.view.hide()
        self.callback(self.view.get_mixture_type())
        return True

    def on_keypress(self, widget, event):
        if event.keyval == Gdk.keyval_from_name("Escape"):
            self.view.hide()
            return True
        if event.keyval == Gdk.keyval_from_name("Return"):
            return self.on_btn_ok_clicked(event)

    def on_window_edit_dialog_delete_event(self, event, args=None):
        self.view.hide()
        return True # do not propagate

    pass # end of class


#!/usr/bin/python

# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import unittest

from test.utils import create_object_attribute_test

from pyxrd.phases.models.unit_cell_prop import UnitCellProperty

__all__ = [
    'TestPhase',
]

class TestPhase(unittest.TestCase):

    phase = None

    def setUp(self):
        self.ucp = UnitCellProperty(
            name="TestUCP",
            value=0.0,
            enabled=False,
            factor=0.0,
            constant=0.0,
            prop=None,
            parent=None
        )

    def tearDown(self):
        del self.ucp

    def test_not_none(self):
        self.assertIsNotNone(self.ucp)

    def test_value_of_prop(self):
        class Dummy():
            attribute = "Test123"

        dummy = Dummy()
        self.ucp.prop = (dummy, "attribute")
        self.assertEqual(self.ucp.get_value_of_prop(), dummy.attribute)

        self.ucp.prop = (None, "attribute")
        self.assertEqual(self.ucp.get_value_of_prop(), 0.0)

    def test_update_value(self):
        class Dummy():
            attribute = 0.5

        dummy = Dummy()
        self.ucp.prop = (dummy, "attribute")
        self.ucp.factor = 0.5
        self.ucp.constant = 1.0

        #Check that if the prop is disabled, it can be set manually:
        self.ucp.value = 0.075
        self.assertEqual(self.ucp.value, 0.075)

        #Check that if the prop is enabled, it is calculated automatically:
        self.ucp.enabled = True
        self.assertEqual(self.ucp.value, 0.5 * 0.5 + 1.0)

        #Check that if the prop is enabled, it can't be set manually:
        self.ucp.value = 0.075
        self.assertNotEqual(self.ucp.value, 0.075)

    test_name = create_object_attribute_test("ucp", "name", "Test Name")
    test_name = create_object_attribute_test("ucp", "value", 0.5)
    test_name = create_object_attribute_test("ucp", "factor", 0.5)
    test_name = create_object_attribute_test("ucp", "constant", 0.5)
    test_name = create_object_attribute_test("ucp", "prop", (None, ""))
    test_name = create_object_attribute_test("ucp", "enabled", True)
    test_name = create_object_attribute_test("ucp", "inherited", True)

    pass # end of class



#!/usr/bin/python

# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from pyxrd.project.models import Project

def run(args):
    project = Project(name="Test")
    project.name = "Test123"
    pass



#!/usr/bin/python

# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import unittest

from test.utils import create_object_attribute_test

from pyxrd.phases.models.CSDS import LogNormalCSDSDistribution, DritsCSDSDistribution

__all__ = [
    'TestLogNormalCSDSDistribution',
    'TestDritsCSDSDistribution'
]

class TestLogNormalCSDSDistribution(unittest.TestCase):

    component = None

    def setUp(self):
        self.CSDS = LogNormalCSDSDistribution()

    def tearDown(self):
        del self.CSDS

    def test_not_none(self):
        self.assertIsNotNone(self.CSDS)

    def test_data_object(self):
        self.assertIsNotNone(self.CSDS.data_object)

    test_average = create_object_attribute_test("CSDS", "average", 15)
    test_alpha_scale = create_object_attribute_test("CSDS", "alpha_scale", 0.5)
    test_alpha_offset = create_object_attribute_test("CSDS", "alpha_offset", 0.6)
    test_beta_scale = create_object_attribute_test("CSDS", "alpha_scale", 0.5)
    test_beta_offset = create_object_attribute_test("CSDS", "alpha_offset", 0.6)

    pass # end of class

class TestDritsCSDSDistribution(unittest.TestCase):

    component = None

    def setUp(self):
        self.CSDS = DritsCSDSDistribution()

    def tearDown(self):
        del self.CSDS

    def test_not_none(self):
        self.assertIsNotNone(self.CSDS)

    def test_data_object(self):
        self.assertIsNotNone(self.CSDS.data_object)

    test_average = create_object_attribute_test("CSDS", "average", 15)

    pass # end of class


# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from pkg_resources import resource_filename # @UnresolvedImport

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

from pyxrd.generic.views import DialogView, BaseView

class CalculatedLinePropertiesView(BaseView):
    builder = resource_filename(__name__, "glade/lines/calculated_props.glade")
    top = "cal_line_props"
    widget_format = "cal_%s"

    widget_groups = {
        'full_mode_only': [
            "cal_line_props"
        ]
    }

    pass # end of class

class ExperimentalLinePropertiesView(BaseView):
    builder = resource_filename(__name__, "glade/lines/experimental_props.glade")
    top = "exp_line_props"
    widget_format = "exp_%s"

    pass # end of class

class BackgroundView(DialogView):
    title = "Remove Background"
    subview_builder = resource_filename(__name__, "glade/lines/background.glade")
    subview_toplevel = "edit_background"
    modal = True
    resizable = False

    def_bg_view = "bg_linear"
    bg_view_cont = "bg_view_container"

    def select_bg_view(self, bg_view=None):
        if bg_view is not None:
            bg_view = "bg_%s" % bg_view
        else:
            bg_view = self.def_bg_view
        self._add_child_view(self[bg_view], self[self.bg_view_cont])

    def set_file_dialog(self, dialog, callback):
        fcb_bg_pattern = Gtk.FileChooserButton(dialog)
        fcb_bg_pattern.connect("file-set", callback, dialog)
        self["fcb_bg_container"].add(fcb_bg_pattern)

    pass #end of class

class AddNoiseView(DialogView):
    title = "Add Noise"
    subview_builder = resource_filename(__name__, "glade/lines/add_noise.glade")
    subview_toplevel = "add_noise"
    modal = True
    resizable = False

    pass # end of class

class SmoothDataView(DialogView):
    title = "Smooth Data"
    subview_builder = resource_filename(__name__, "glade/lines/smoothing.glade")
    subview_toplevel = "smooth_data"
    modal = True
    resizable = False

    pass # end of class

class ShiftDataView(DialogView):
    title = "Shift Pattern"
    subview_builder = resource_filename(__name__, "glade/lines/shifting.glade")
    subview_toplevel = "shift_pattern"
    modal = True
    resizable = False

    pass # end of class

class StripPeakView(DialogView):
    title = "Strip Peak"
    subview_builder = resource_filename(__name__, "glade/lines/strip_peak.glade")
    subview_toplevel = "strip_peak"
    modal = True
    resizable = False

    pass # end of class

class CalculatePeakPropertiesView(DialogView):
    title = "Calculate Peak Properties"
    subview_builder = resource_filename(__name__, "glade/lines/peak_properties.glade")
    subview_toplevel = "peak_properties"
    modal = True
    resizable = False

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from pyxrd.generic.io.utils import get_case_insensitive_glob

from pyxrd.file_parsers.registry import ParserNamespace
from pyxrd.file_parsers.xrd_parsers.csv_parser import GenericXYCSVParser

exc_parsers = ParserNamespace("EXCParsers")

@exc_parsers.register_parser()
class EXCParser(GenericXYCSVParser):
    """
        ASCII *.DAT, *.CSV and *.TAB format parser
    """

    description = "Exclusion range file"
    extensions = get_case_insensitive_glob("*.EXC")

    pass # end of class


#!/usr/bin/python

# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from .base import AbstractTestProbModel

from test.utils import create_object_attribute_test

from pyxrd.probabilities.models.R0models import *
import unittest

__all__ = [
    'TestR0G1Model',
    'TestR0G2Model',
    'TestR0G3Model',
    'TestR0G4Model',
    'TestR0G5Model',
    'TestR0G6Model'
]

class TestR0G1Model(AbstractTestProbModel, unittest.TestCase):

    prob_model_type = R0G1Model

    pass # end of class

class TestR0G2Model(AbstractTestProbModel, unittest.TestCase):

    prob_model_type = R0G2Model

    test_F1 = create_object_attribute_test("prob_model", "F1", 0.7)

    pass # end of class

class TestR0G3Model(AbstractTestProbModel, unittest.TestCase):

    prob_model_type = R0G3Model

    test_F1 = create_object_attribute_test("prob_model", "F1", 0.7)
    test_F2 = create_object_attribute_test("prob_model", "F2", 0.7)

    pass # end of class

class TestR0G4Model(AbstractTestProbModel, unittest.TestCase):

    prob_model_type = R0G4Model

    test_F1 = create_object_attribute_test("prob_model", "F1", 0.7)
    test_F2 = create_object_attribute_test("prob_model", "F2", 0.7)
    test_F3 = create_object_attribute_test("prob_model", "F3", 0.7)

    pass # end of class

class TestR0G5Model(AbstractTestProbModel, unittest.TestCase):

    prob_model_type = R0G5Model

    test_F1 = create_object_attribute_test("prob_model", "F1", 0.7)
    test_F2 = create_object_attribute_test("prob_model", "F2", 0.7)
    test_F3 = create_object_attribute_test("prob_model", "F3", 0.7)
    test_F4 = create_object_attribute_test("prob_model", "F4", 0.7)


    pass # end of class

class TestR0G6Model(AbstractTestProbModel, unittest.TestCase):

    prob_model_type = R0G6Model

    test_F1 = create_object_attribute_test("prob_model", "F1", 0.7)
    test_F2 = create_object_attribute_test("prob_model", "F2", 0.7)
    test_F3 = create_object_attribute_test("prob_model", "F3", 0.7)
    test_F4 = create_object_attribute_test("prob_model", "F4", 0.7)
    test_F5 = create_object_attribute_test("prob_model", "F5", 0.7)

    pass # end of class



#!/usr/bin/python

# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon

from .dummy_async_result import DummyAsyncResult

class DummyAsyncServer(object):
    """ A non-asynchronous dummy implementation of the AsyncServer object """
    
    def loopCondition(self):
        return True

    def submit(self, func):
        return DummyAsyncResult(func)

    def shutdown(self):
        pass
    
    pass #end of class


# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from contextlib import contextmanager
from os.path import dirname

import numpy as np

from mvc import Controller
from mvc.adapters.gtk_support.dialogs.dialog_factory import DialogFactory

from pyxrd.generic.controllers import BaseController, ObjectListStoreController
from pyxrd.atoms.models import AtomType
from pyxrd.atoms.views import EditAtomTypeView
from pyxrd.data import settings
from pyxrd.file_parsers.atom_type_parsers import atom_type_parsers

class EditAtomTypeController(BaseController):
    """
        The controller for the AtomType model and EditAtomTypeView view.
    """

    # ------------------------------------------------------------
    #      Initialization and other internals
    # ------------------------------------------------------------
    def register_adapters(self):
        self.update_plot()

    # ------------------------------------------------------------
    #      Methods & Functions
    # ------------------------------------------------------------
    def update_plot(self):
        x, y = (), ()
        if self.model is not None:
            x = np.arange(0, 90.0, 90.0 / 100.0)
            y = self.model.get_atomic_scattering_factors(2 * np.sin(np.radians(x / 2)) / settings.DEFAULT_LAMBDA)
        self.view.update_figure(x, y)

    # ------------------------------------------------------------
    #      Notifications of observable properties
    # ------------------------------------------------------------
    @Controller.observe("parameters_changed", signal=True)
    def notif_parameter_changed(self, model, prop_name, info):
        self.update_plot()

    pass # end of class

class AtomTypesController(ObjectListStoreController): # FIXME THIS NEES CLEAN-UP AND TESTING!!
    """
        Controller for an AtomType ObjectListStore model and view.
    """
    file_filters = property(fget=lambda *a: atom_type_parsers.get_file_filters())
    export_filters = property(fget=lambda *a: atom_type_parsers.get_export_file_filters())
    treemodel_property_name = "atom_types"
    treemodel_class_type = AtomType
    columns = [ ("Atom type name", "c_name") ]
    delete_msg = "Deleting an atom type is irreversible!\nAre You sure you want to continue?"
    obj_type_map = [
        (AtomType, EditAtomTypeView, EditAtomTypeController),
    ]
    title = "Edit Atom Types"

    _export_atomtypes_dialog = None
    @property
    def export_atomtypes_dialog(self):
        """ Creates & returns the 'export atom types' dialog """
        if self._export_atomtypes_dialog is None:
            # Default location of the database:
            current_folder = dirname(settings.DATA_REG.get_file_path("ATOM_SCAT_FACTORS"))
            # Create the dialog once, and re-use
            self._export_atomtypes_dialog = DialogFactory.get_save_dialog(
                title="Export atom types",
                current_folder=current_folder,
                persist=True,
                filters=self.export_filters,
                parent=self.view.get_top_widget()
            )
        return self._export_atomtypes_dialog

    _import_atomtypes_dialog = None
    @property
    def import_atomtypes_dialog(self):
        """ Creates & returns the 'import atom types' dialog """
        if self._import_atomtypes_dialog is None:
            # Default location of the database:
            current_folder = dirname(settings.DATA_REG.get_file_path("ATOM_SCAT_FACTORS"))
            # Create the dialog once, and re-use
            self._import_atomtypes_dialog = DialogFactory.get_load_dialog(
                title="Import atom types",
                current_folder=current_folder,
                persist=True, multiple=False,
                filters=self.file_filters,
                parent=self.view.get_top_widget()
            )
        return self._import_atomtypes_dialog

    # ------------------------------------------------------------
    #      Methods & Functions
    # ------------------------------------------------------------
    def get_atom_types_tree_model(self, *args):
        return self.treemodel

    def create_new_object_proxy(self):
        return AtomType(name="New Atom Type", parent=self.model)

    @contextmanager
    def _multi_operation_context(self):
        with self.model.data_changed.hold():
            yield

    # ------------------------------------------------------------
    #      GTK Signal handlers
    # ------------------------------------------------------------
    def on_load_object_clicked(self, event):
        self.import_atomtypes_dialog.run(
            lambda dialog: self.model.load_atom_types(
                dialog.filename,
                dialog.parser
            )
        )

    def on_save_object_clicked(self, event):
        self.export_atomtypes_dialog.run(
            lambda dialog: dialog.parser.write(
                dialog.filename,
                self.get_selected_objects(),
                AtomType.Meta.get_local_persistent_properties()
            )
        )

    pass # end of class




# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import argparse, os
from pyxrd.__version import __version__
from .appdirs import user_data_dir, user_log_dir

### General Information ###
VERSION = __version__

DEBUG = False
FINALIZERS = [] #A list of callables that are called before the main function is left
BGSHIFT = True

LOG_FILENAME = os.path.join(user_log_dir('PyXRD'), 'errors.log')

### The URL where PyXRD looks for updates & the online manual ###
UPDATE_URL = 'http://users.ugent.be/~madumon/pyxrd/'
MANUAL_URL = UPDATE_URL

### Factor to multiply the CSDS average with to obtain the maximum CSDS ###
LOG_NORMAL_MAX_CSDS_FACTOR = 2.5

### What type of residual error we use: ###
#   "Rp" = 'standard' pattern Rp factor
#   "Rpw" = 'weighted' pattern Rp factor
#   "Rpder" = Rp factor of first-derivatives
RESIDUAL_METHOD = "Rp"

### Default wavelength if no Goniometer is available ###
DEFAULT_LAMBDA = 0.154056

### GUI Mode (for HPC turn to False) ###
GUI_MODE = True


### Size of parameter space record in bytes ###
PAR_SPACE_MEMORY_LIMIT = 25 * (1024 * 1024)

### Default Styles & Colors ###
DEFAULT_LAYOUT = "VIEWER" # one of the keys in DEFAULT_LAYOUTS!
DEFAULT_LAYOUTS = {
    "FULL": "Full",
    "VIEWER": "View-mode"
}

AXES_MANUAL_XMIN = 0.0
AXES_MANUAL_XMAX = 70.0
AXES_XSTRETCH = False
AXES_DSPACING = False
AXES_DEFAULT_WAVELENGTH = 0.154056

AXES_XLIMIT = 0
AXES_XLIMITS = {
    0: "Automatic",
    1: "Manual"
}

AXES_MANUAL_YMIN = 0.0
AXES_MANUAL_YMAX = 0.0
AXES_YVISIBLE = False

AXES_YNORMALIZE = 0
AXES_YNORMALIZERS = {
    0: "Multi normalised",
    1: "Single normalised",
    2: "Unchanged raw counts",
}

AXES_YLIMIT = 0
AXES_YLIMITS = {
    0: "Automatic",
    1: "Manual"
}

EXPERIMENTAL_COLOR = "#000000"
CALCULATED_COLOR = "#FF0000"

EXPERIMENTAL_LINEWIDTH = 1.0
CALCULATED_LINEWIDTH = 2.0

CALCULATED_LINESTYLE = "-"
EXPERIMENTAL_LINESTYLE = "-"

CALCULATED_MARKER = ""
EXPERIMENTAL_MARKER = ""

PATTERN_LINE_STYLES = {
    "": "Nothing",
    "-": "Solid",
    "--": "Dashed",
    "-.": "Dash Dot",
    ":": "Dotted"
}

PATTERN_MARKERS = {
    "": "No marker",
    ".": "Point",
    ",": "Pixel",
    "+": "Plus",
    "x": "Cross",
    "D": "Diamond",
    "o": "Circle",
    "v": "Triangle down",
    "^": "Triangle up",
    "<": "Triangle left",
    ">": "Triangle right",
    "8": "Octagon",
    "s": "Square",
    "p": "Pentagon",
    "*": "Star",
    "h": "Hexagon",
}

PATTERN_SHIFT_POSITIONS = {
    0.42574: "Quartz    0.42574   SiO2",
    0.3134:  "Silicon   0.31355   Si",
    0.2476:  "Zincite   0.24759   ZnO",
    0.2085:  "Corundum  0.2085    Al2O3",
    0.4183:  "Goethite  0.4183    FeO(OH)",
    0.48486: "Gibbsite  0.48486   Al(OH)3",
}
PATTERN_SHIFT_TYPE = "Displacement" # or "Linear"
PATTERN_SMOOTH_TYPES = { 0: "Moving Triangle" }
PATTERN_BG_TYPES = { 0: "Linear", 1: "Pattern" }

DIVERGENCE_MODES = {
    "AUTOMATIC": "Automatic divergence",
    "FIXED": "Fixed divergence"
}
DEFAULT_DIVERGENCE_MODE = "FIXED"
DEFAULT_SAMPLE_LENGTH = 1.25 # in cm

PLOT_OFFSET = 0.75
PATTERN_GROUP_BY = 1
LABEL_POSITION = 0.35

MARKER_VISIBLE = True
MARKER_X_OFFSET = 0.0
MARKER_Y_OFFSET = 0.05
MARKER_POSITION = 0.0

MARKER_INHERIT_COLOR = True
MARKER_COLOR = "#000000"
MARKER_INHERIT_ANGLE = True
MARKER_ANGLE = 0.0
MARKER_INHERIT_TOP_OFFSET = True
MARKER_TOP_OFFSET = 0.0
MARKER_INHERIT_BASE = True
MARKER_BASE = 1
MARKER_BASES = {
    0: "X-axis",
    1: "Experimental profile",
    2: "Calculated profile",
    3: "Lowest of both",
    4: "Highest of both"
}
MARKER_INHERIT_TOP = True
MARKER_TOP = 0
MARKER_TOPS = {
     0: "Relative to base",
     1: "Top of plot"
}
MARKER_INHERIT_STYLE = True
MARKER_STYLE = "none"
MARKER_STYLES = {
    "none": "None", "solid": "Solid",
    "dashed": "Dash", "dotted": "Dotted",
    "dashdot": "Dash-Dotted", "offset": "Display at Y-offset"
}
MARKER_INHERIT_ALIGN = True
MARKER_ALIGN = "left"
MARKER_ALIGNS = {
    "left": "Left align",
    "center": "Centered",
    "right": "Right align"
}

EXCLUSION_FOREG = "#999999"
EXCLUSION_LINES = "#333333"

### Plot Information ###
PLOT_TOP = 0.85
MAX_PLOT_RIGHT = 0.95
PLOT_BOTTOM = 0.0
PLOT_LEFT = 0.05
PLOT_HEIGHT = PLOT_TOP - PLOT_BOTTOM

OUTPUT_PRESETS = [
    ("Landscape Large print", 8000, 4800, 300.0),
    ("Landscape Medium print", 6000, 3800, 300.0),
    ("Landscape Small print", 4000, 2800, 300.0),
    ("Portrait Large print", 4800, 8000, 300.0),
    ("Portrait Medium print", 3800, 6000, 300.0),
    ("Portrait Small print", 2800, 4000, 300.0),
]

### Default Directories & Files ###
DATA_REG = None # set at run-time
DATA_DIRS = [
    ("DEFAULT_DATA", "./", None),
    ("USER_DATA", user_data_dir('PyXRD'), None),
    ("LOG_DIR", user_log_dir('PyXRD'), None),
    ("DEFAULT_PHASES", "default phases/", "USER_DATA"),
    ("DEFAULT_COMPONENTS", "default components/", "DEFAULT_DATA"),
    ("DEFAULT_GONIOS", "default goniometers/", "DEFAULT_DATA"),
    ("DEFAULT_WL_DISTR", "default wavelength distributions/", "DEFAULT_DATA"),
    ("APPLICATION_ICONS", "icons/", "DEFAULT_DATA"),
]
DATA_FILES = [
    ("COMPOSITION_CONV", "composition_conversion.csv", "DEFAULT_DATA"),
    ("ATOM_SCAT_FACTORS", "atomic scattering factors.atl", "DEFAULT_DATA"),
    ("MINERALS", "mineral_references.csv", "DEFAULT_DATA"),
]

### Async calculation providers ###
ASYNC_SERER_PROVIDERS = [
   "pyxrd.server.provider.Pyro4AsyncServerProvider",
   "pyxrd.generic.asynchronous.dummy_async_provider.DummyAsyncServerProvider",
]
ASYNC_SERVER_PRELOAD = True

### Runtime Settings Retrieval ###
SETTINGS_APPLIED = False
ARGS = None

def _parse_args():
    """ Parses command line arguments """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "filename", nargs="?", default="",
        help="A PyXRD project filename"
    )
    parser.add_argument(
        "-s", "--script", default="",
        help="Can be used to pass a script containing a run() function"
    )
    parser.add_argument(
        "-d", "--debug", dest='debug', action='store_const',
        const=True, default=False,
        help='Run in debug mode'
    )

    args = parser.parse_args()
    del parser # free some memory
    return args

__apply_lock__ = False
def initialize(override_debug=DEBUG):
    """Apply runtime settings, can and needs to be called only once"""
    global __apply_lock__, SETTINGS_APPLIED
    if not __apply_lock__ and not SETTINGS_APPLIED:
        __apply_lock__ = True

        # Get command line arguments
        global ARGS
        ARGS = _parse_args()

        # Set gui flag
        global GUI_MODE
        GUI_MODE = not bool(ARGS.script)

        # Set debug flag
        global DEBUG
        DEBUG = ARGS.debug or override_debug

        # Setup data registry:
        global DATA_REG, DATA_DIRS, DATA_FILES
        from pyxrd.generic.io.data_registry import DataRegistry
        DATA_REG = DataRegistry(dirs=DATA_DIRS, files=DATA_FILES)

        # If we are running in GUI mode, setup GUI stuff:
        if GUI_MODE:
            import matplotlib
            import gi
            gi.require_version('Gtk', '3.0')
            from gi.repository import Gtk, GdkPixbuf

            # Setup matplotlib fonts:
            font = {
                'weight' : 'heavy', 'size': 14,
		'sans-serif' : 'Helvetica, Arial, sans-serif',
                'family' : 'sans-serif',
            }
            matplotlib.rc('font', **font)
            mathtext = {
                'default': 'regular',
                'fontset': 'stixsans',
            }
            matplotlib.rc('mathtext', **mathtext)
            # matplotlib.rc('text', **{'usetex':True})
            # Load our own icons:
            iconfactory = Gtk.IconFactory()
            icons_path = DATA_REG.get_directory_path("APPLICATION_ICONS")
            for root, dirnames, filenames in os.walk(icons_path):
                for filename in filenames:
                    if filename.endswith(".png"):
                        stock_id = filename[:-4] # remove extensions
                        filename = "%s/%s" % (icons_path, filename)                        
                        pixbuf = GdkPixbuf.Pixbuf.new_from_file(filename)
                        iconset = Gtk.IconSet(pixbuf)
                        iconfactory.add(stock_id, iconset)
            iconfactory.add_default()

        # Make sure default directories exist:
        for path in DATA_REG.get_all_directories():
            try:
                os.makedirs(path)
            except OSError:
                pass

        # Free some memory at this point:
        import gc
        gc.collect()

        # Log that we did all of this:
        import logging
        logger = logging.getLogger(__name__)
        logger.info("Runtime settings applied")

    __apply_lock__ = False
    SETTINGS_APPLIED = True

# ## end of settings



# coding=UTF-8
# ex:ts=4:sw=4:et

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from .event_delegator import MPLCanvasEventDelegator

class ClickCatcher():
    """
        This class can be used to register matplotlib artists, which will
        fire the update_callback method when clicked. When registering
        the artist, an arbitrary object can be passed which is passed to the
        callback. 
    """
    def __init__(self, plot_controller, update_callback=None):
        self.plot_controller = plot_controller
        self._canvas = plot_controller.canvas
        self._window = self._canvas.get_window()
        self._update_callback = update_callback
        self.connect()
        self._artists = {}

    def register_artist(self, artist, obj):
        self._artists[artist] = obj
        artist.set_picker(True)
   
    def _on_pick(self, event):
        if event.artist is not None:
            obj = self._artists.get(event.artist, None)
            self._update_callback(obj)
        return False

    def connect(self):
        delegator = MPLCanvasEventDelegator.wrap_canvas(self._canvas)
        delegator.connect('pick_event', self._on_pick, first=True)

    def disconnect(self):
        delegator = MPLCanvasEventDelegator.wrap_canvas(self._canvas)
        delegator.disconnect('pick_event', self._on_pick)

    pass #end of class


# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from io import StringIO

from scipy.optimize import fmin_l_bfgs_b

from .exceptions import wrap_exceptions

def setup_project(projectf):
    from pyxrd.file_parsers.json_parser import JSONParser
    from pyxrd.project.models import Project
    type(Project).object_pool.clear()

    f = StringIO(projectf)
    project = JSONParser.parse(f)
    f.close()

    return project

@wrap_exceptions
def run_refinement(projectf, mixture_index):
    """
        Runs a refinement setup for 
            - projectf: project data
            - mixture_index: what mixture in the project to use
    """
    if projectf is not None:
        from pyxrd.data import settings
        settings.initialize()

        # Retrieve project and mixture:
        project = setup_project(projectf)
        del projectf

        import gc
        gc.collect()

        mixture = project.mixtures[mixture_index]
        mixture.refinement.update_refinement_treestore()
        refiner = mixture.refinement.get_refiner()
        refiner.refine()

        return list(refiner.history.best_solution), refiner.history.best_residual

@wrap_exceptions
def improve_solution(projectf, mixture_index, solution, residual, l_bfgs_b_kwargs={}):
    if projectf is not None:
        from pyxrd.data import settings
        settings.initialize()

        # Retrieve project and mixture:
        project = setup_project(projectf)
        del projectf

        mixture = project.mixtures[mixture_index]

        with mixture.data_changed.ignore():

            # Setup context again:
            mixture.update_refinement_treestore()
            refiner = mixture.refinement.get_refiner()

            # Refine solution
            vals = fmin_l_bfgs_b(
                refiner.get_residual,
                solution,
                approx_grad=True,
                bounds=refiner.ranges,
                **l_bfgs_b_kwargs
            )
            new_solution, new_residual = tuple(vals[0:2])

        # Return result
        return new_solution, new_residual
    else:
        return solution, residual



# coding=UTF-8
# ex:ts=4:sw=4:et

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from .event_delegator import MPLCanvasEventDelegator

class MotionTracker():
    """
        A wrapper that tracks mouse movements on a plot. Will call
        the update_callback with an x position and the click event object as
        arguments.
    """
    def __init__(self, plot_controller, update_callback=None):
        self._canvas = plot_controller.canvas
        self._window = self._canvas.get_window()
        self._update_callback = update_callback
        self.connect()

    def _on_motion(self, event):
        x_pos = -1
        if event.inaxes:
            x_pos = event.xdata
        if callable(self._update_callback):
            self._update_callback(x_pos, event)
        return False

    def connect(self):
        delegator = MPLCanvasEventDelegator.wrap_canvas(self._canvas)
        delegator.connect('motion_notify_event', self._on_motion, first=True)

    def disconnect(self):
        delegator = MPLCanvasEventDelegator.wrap_canvas(self._canvas)
        delegator.disconnect('motion_notify_event', self._on_motion)

    pass #end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gdk

from pyxrd.generic.controllers.base_controller import BaseController

class DialogController(BaseController):
    """
        Simple controller which has a DialogView subclass instance as view.
    """

    # ------------------------------------------------------------
    #      GTK Signal handlers
    # ------------------------------------------------------------
    def on_btn_ok_clicked(self, event):
        self.on_cancel()
        return True

    def on_keypress(self, widget, event):
        if event.keyval == Gdk.keyval_from_name("Escape"):
            self.on_cancel()
            return True

    def on_window_edit_dialog_delete_event(self, event, args=None):
        self.on_cancel()
        return True # do not propagate

    def on_cancel(self):
        self.view.hide()

    pass #end of class


#!/usr/bin/python

# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import os, sys
from pyxrd.project.models import Project


def run(args):
    # generates a project file containing the phases as described by the Sybilla XML output:
    if args and args.filename != "":
        # Import:
        project = Project.create_from_sybilla_xml(args.filename)

        # Save this right away:
        project_filename = "%s/%s" % (os.path.dirname(args.filename), os.path.basename(args.filename).replace(".xml", ".pyxrd", 1))

        from pyxrd.file_parsers.json_parser import JSONParser
        JSONParser.write(project, project_filename, zipped=True)

        # Relaunch processs
        args = [sys.argv[0], project_filename, ]
        args.insert(0, sys.executable)
        if sys.platform == 'win32':
            args = ['"%s"' % arg for arg in args]
        os.execv(sys.executable, args)
        sys.exit(0)




# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from mvc.models.properties import BoolProperty, FloatProperty

from pyxrd.generic.mathtext_support import mt_range
from pyxrd.generic.io import storables
from pyxrd.generic.utils import not_none
from pyxrd.generic.models.properties import InheritableMixin

from pyxrd.refinement.refinables.properties import RefinableMixin

from .base_models import _AbstractProbability
from mvc.models.properties.action_mixins import SetActionMixin

__all__ = [
    "R3G2Model"
]

@storables.register()
class R3G2Model(_AbstractProbability):
    r"""
    (Restricted) probability model for Reichweite 3 with 2 components.
    
    The (due to restrictions only) 2 independent variables are:
    
    .. math::
        :nowrap:
        
        \begin{align*}
            & W_1
            & P_{1111} \text{(if $\nicefrac{2}{3} \leq W_1 < \nicefrac{3}{4}$) or} P_{2112} \text{(if $\nicefrac{3}{4} \leq W_1 \leq 1)$} \\
        \end{align*}
    
    This model can only describe mixed layers with more than 
    :math:`\nicefrac{2}{3}` of the layers being of the first type, no two layers
    of the second type occur after each other, and in which the probability of
    finding a layer of the first type in between two layers of the second type
    is zero. This translates to the following conditions:
    
    .. math::
        :nowrap:
        
        \begin{align*}
	        & \nicefrac{2}{3} <= W_1 <= 1 \\
	        & P_{22} = 0  \\
	        & P_{212} = 0 \\
	        & \\
	        & \text{Since $P_{22} = 0$ and $P_{212} = 0$:} \\
            & P_{1122} = P_{1212} = 0 \\
            & \text{And thus:} \\
            & P_{1121} = P_{1211} = 1 \\
        \end{align*} 
        
    The following probabilities are undefined, but are set to zero or one to
    make the validation correct. This doesn't matter much, since the weight
    fractions these probabilities are multiplied with, equal zero anyway
    (e.g. :math:`W_{2211} = W_{22} * P_{221} * P_{2211}` and :math:`W_{22}` 
    is zero since :math:`P_{22}` is zero):
        
    .. math::
        :nowrap:
        
        \begin{align*}
            & P_{1121} &= P_{1211} &= P_{2211} 
            &= P_{2121} &= P_{2221} &= P_{1221} = 0 \\
            & P_{1122} &= P_{1212} &= P_{2212} 
            &= P_{2122} &= P_{2222} &= P_{1222} = 1 \\
        \end{align*}  

    The remaining probabilities and weight fractions can be calculated as 
    follows:
    
   .. math::
        :nowrap:
        
        \begin{align*}
            & W_2 = 1 - W_1 \\
            & \\
            & \text{if $W_1 < \nicefrac{3}{4}$:} \\
            & \quad \text{$P_{1111}$ is given}\\
            & \quad P_{1112} = 1 - P_{1111} \\
            & \quad P_{2111} = P_{1112} * \frac{W_1 - 2 \cdot W_2}{W_2} \\
            & \quad P_{2112} = 1 - P_{2111} \\  
            & \\
            & \text{if $W_1 \geq \nicefrac{3}{4}$:} \\
            & \quad \text{$P_{2112}$ is given}\\
            & \quad P_{2111} = 1 - P_{2112} \\
            & \quad P_{1111} = P_{2111} * \frac{W_2}{W_1 - 2 \cdot W_2} \\
            & \quad P_{1112} = 1 - P_{1111} \\  
            & \\
            & W_{111} = 3 \cdot W_1 - 2 \\
            & W_{212} = W_{221} = W_{222} = W_{122} = 0 \\
            & W_{211} = W_{121} = W_{112} = 1 - W_1 \\
        \end{align*}
	
	"""

    # MODEL METADATA:
    class Meta(_AbstractProbability.Meta):
        store_id = "R3G2Model"

    # PROPERTIES:
    _G = 2

    inherit_W1 = BoolProperty(
        default=False, text="Inherit flag for W1",
        persistent=True, visible=True,
        set_action_name="update",
        mix_with=(SetActionMixin,)
    )
    W1 = FloatProperty(
        default=0.85, text="W1 (> 2/3)", math_text=r"$W_1 (> \frac{2}{3})$",
        persistent=True, visible=True, refinable=True, store_private=True,
        minimum=2.0 / 3.0, maximum=1.0, is_independent=True, inheritable=True,
        inherit_flag="inherit_W1", inherit_from="parent.based_on.probabilities.W1",
        set_action_name="update",
        mix_with=(SetActionMixin, RefinableMixin, InheritableMixin)
    )

    inherit_P1111_or_P2112 = BoolProperty(
        default=False, text="Inherit flag for P1111_or_P2112",
        persistent=True, visible=True,
        set_action_name="update",
        mix_with=(SetActionMixin,)
    )
    P1111_or_P2112 = FloatProperty(
        default=0.75, text="P1111 (W1 < 3/4) or\nP2112 (W1 > 3/4)",
        math_text=r"$P_{1111} %s$ or $\newline P_{2112} %s$" % (
            mt_range(2.0 / 3.0, "W_1", 3.0 / 4.0),
            mt_range(3.0 / 4.0, "W_1", 1.0)),
        persistent=True, visible=True, refinable=True, store_private=True,
        minimum=0.0, maximum=1.0, is_independent=True, inheritable=True,
        inherit_flag="inherit_P1111_or_P2112", inherit_from="parent.based_on.probabilities.P1111_or_P2112",
        set_action_name="update",
        mix_with=(SetActionMixin, RefinableMixin, InheritableMixin)
    )

    # ------------------------------------------------------------
    #      Initialization and other internals
    # ------------------------------------------------------------
    def __init__(self, W1=0.85, P1111_or_P2112=0.75, *args, **kwargs):
        super(R3G2Model, self).__init__(R=3, *args, **kwargs)

        with self.data_changed.hold():
            self.W1 = not_none(W1, 0.85)
            self.P1111_or_P2112 = not_none(P1111_or_P2112, 0.75)

            self.update()

    # ------------------------------------------------------------
    #      Methods & Functions
    # ------------------------------------------------------------
    def update(self):
        with self.monitor_changes():
            self.mW[0] = self.W1
            self.mW[1] = 1.0 - self.W1

            if self.mW[0] <= 0.75: # 0,0,0,0 is given
                self.mP[0, 0, 0, 0] = self.P1111_or_P2112
                self.mP[0, 0, 0, 1] = max(min(1.0 - self.mP[0, 0, 0, 0], 1.0), 0.0)
                self.mP[1, 0, 0, 0] = max(min(self.mP[0, 0, 0, 1] * (self.mW[0] - 2 * self.mW[1]) / self.mW[1], 1.0), 0.0)
                self.mP[1, 0, 0, 1] = max(min(1.0 - self.mP[1, 0, 0, 0], 1.0), 0.0)
            else: # 1,0,0,1 is given
                self.mP[1, 0, 0, 1] = self.P1111_or_P2112
                self.mP[1, 0, 0, 0] = max(min(1.0 - self.mP[1, 0, 0, 1], 1.0), 0.0)
                self.mP[0, 0, 0, 0] = max(min(1.0 - self.mP[1, 0, 0, 0] * self.mW[1] / (self.mW[0] - 2 * self.mW[1]), 1.0), 0.0)
                self.mP[0, 0, 0, 1] = max(min(1.0 - self.mP[0, 0, 0, 0], 1.0), 0.0)

            # since P11=0 and P101=0, actual values don't matter:
            self.mP[0, 0, 1, 0] = 1.0
            self.mP[0, 0, 1, 1] = 0.0

            self.mP[0, 1, 0, 0] = 1.0
            self.mP[0, 1, 0, 1] = 0.0

            self.mP[0, 1, 1, 0] = 1.0
            self.mP[0, 1, 1, 1] = 0.0

            self.mP[1, 0, 1, 0] = 1.0
            self.mP[1, 0, 1, 1] = 0.0

            self.mP[1, 1, 0, 0] = 1.0
            self.mP[1, 1, 0, 1] = 0.0

            self.mP[1, 1, 1, 0] = 1.0
            self.mP[1, 1, 1, 1] = 0.0

            self.mW[0, 0, 0] = max(min(3 * self.mW[0] - 2, 1.0), 0.0)
            self.mW[1, 0, 1] = self.mW[1, 1, 0] = self.mW[1, 1, 1] = self.mW[0, 1, 1] = 0.0
            self.mW[1, 0, 0] = self.mW[0, 1, 0] = self.mW[0, 0, 1] = max(min(1 - self.mW[0], 1.0), 0.0)

            self.solve()
            self.validate()

    pass # end of class



import threading
import time

class StatusThread(threading.Thread):
    
    def __init__(self, interval, provider):
        """
            @param interval: In seconds: how often to update the status.
        """
        assert(interval > 0.1)
        self.interval = interval
        self.stop_requested = threading.Event()
        self.provider = provider
        threading.Thread.__init__(self)

    def run(self):
        while not self.stop_requested.isSet():
            time.sleep(self.interval)
            self.provider._update_status(self.provider._get_status())
                
    def stop(self):
        self.stop_requested.set()
        self.join()

    pass #end of class


# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from itertools import product
from copy import deepcopy
from contextlib import contextmanager

import numpy as np

from mvc.models.properties import StringProperty, LabeledProperty

from pyxrd.generic.io import Storable
from pyxrd.generic.models import DataModel
from pyxrd.generic.models.properties import IndexProperty

from pyxrd.refinement.refinables.mixins import RefinementGroup
from pyxrd.refinement.refinables.metaclasses import PyXRDRefinableMeta

class _AbstractProbability(RefinementGroup, DataModel, Storable, metaclass=PyXRDRefinableMeta):

    # MODEL INTEL:
    class Meta(DataModel.Meta):
        pass

    phase = property(DataModel.parent.fget, DataModel.parent.fset)

    # PROPERTIES:
    name = StringProperty(default="Probabilities", text="Name")
    W_valid = LabeledProperty(default=None, text="Valid W matrix")
    W_valid_mask = None
    P_valid = LabeledProperty(default=None, text="Valid P matrix")
    P_valid_mask = None

    _R = -1
    @property
    def R(self):
        return self._R

    @property
    def rank(self):
        return self.G ** max(self.R, 1)

    _G = 0
    @property
    def G(self):
        return self._G

    _W = None
    _P = None

    @IndexProperty
    def mP(self, indeces):
        r, ind = self._get_Pxy_from_indeces(indeces)
        return self._lP[r][ind]
    @mP.setter
    def mP(self, indeces, value):
        r, ind = self._get_Pxy_from_indeces(indeces)
        self._lP[r][ind] = value

    @IndexProperty
    def mW(self, indeces):
        r, ind = self._get_Wxy_from_indeces(indeces)
        return self._lW[r][ind]
    @mW.setter
    def mW(self, indeces, value):
        r, ind = self._get_Wxy_from_indeces(indeces)
        self._lW[r][ind] = value

    @property
    def _flags(self):
        return [
            1 if getattr(self, prop.inherit_flag, False) else 0
            for prop in self.Meta.all_properties if getattr(prop, "inheritable", False) 
        ]

    # REFINEMENT GROUP IMPLEMENTATION:
    @property
    def refine_title(self):
        return self.name

    @property
    def refine_descriptor_data(self):
        return dict(
            phase_name=self.phase.name,
            component_name="*"
        )

    # ------------------------------------------------------------
    #      Initialization and other internals
    # ------------------------------------------------------------
    def __init__(self, R=-1, *args, **kwargs):
        super(_AbstractProbability, self).__init__(*args, **kwargs)
        self._R = R
        self._create_matrices()

    def _create_matrices(self):
        """
            Creates a list of matrices for different 'levels' of R:
                e.g. when R=3 with g layers there can be 4 different W matrixes:
                    Wi = gxg matrix
                    Wij = g²xg² matrix
                    Wijk = g³xg³ matrix
                    Wijkl = another g³xg³ matrix (= Wijk * Pijkl)
                and 3 different P matrices:
                    Pij = gxg matrix
                    Pijk = g²xg² matrix
                    Pijkl = g³xg³ matrix
        """
        R = max(self.R, 1)
        self._lW = [None] * (R + 1)
        self._lP = [None] * R
        for r in range(R):
            lrank = self.G ** (r + 1)
            self._lW[r] = np.zeros(shape=(lrank, lrank), dtype=float)
            self._lP[r] = np.zeros(shape=(lrank, lrank), dtype=float)
        self._lW[-1] = np.zeros(shape=(lrank, lrank), dtype=float)
        self._W = self._lW[-2]
        self._P = self._lP[-1]

        # validity matrices:
        self.W_valid = np.array([False] * (R + 1))
        self.W_valid_mask = np.array([None] * (R + 1))
        self.P_valid = np.array([False] * R)
        self.P_valid_mask = np.array([None] * R)

    # ------------------------------------------------------------
    #      Methods & Functions
    # ------------------------------------------------------------
    def update(self):
        raise NotImplementedError

    def _clamp_set_and_update(self, name, value, minimum=0.0, maximum=1.0):
        clamped = min(max(value, minimum), maximum)
        if getattr(self, name) != clamped:
            setattr(self, name, clamped)
            self.update()

    def solve(self):
        """
            This 'solves' the other W and P matrices using the 'top' P and W
            matrix calculated in the update method.
        """

        for num in range(1, self.R):
            # W matrices:
            for base in product(list(range(self.G)), repeat=num):
                self.mW[base] = 0
                for i in range(self.G):
                    self.mW[base] += self.mW[(i,) + base]
            # P matrices:
            p_num = num + 1
            for base in product(list(range(self.G)), repeat=p_num):
                W = self.mW[base[:-1]]
                self.mP[base] = self.mW[base] / W if W > 0 else 0.0

        # one extra W matrix:
        self._lW[-1][:] = np.dot(self._W, self._P)

    def validate(self):
        """
            Checks wether the calculated matrices are valid, and stores the
            validation results in 'masks': matrices of the same size, in which
            the values correspond with 1 minus the number of validation rules
            that specific W or P value has failed for.
        """
        def _validate_WW(W, R):
            """Validation rules for the product of a W and P matrix"""
            W_valid_mask = np.ones_like(W)
            rank = self.G ** max(R, 1)

            # sum of the cols (W...x's) need to equal W...
            for i in range(rank):
                if abs(np.sum(W[..., i]) - self._W[i, i]) > 1e4:
                    W_valid_mask[..., i] -= 1

            # sum of the entire matrix must equal one:
            if abs(np.sum(W) - 1.0) > 1e4:
                W_valid_mask -= 1

            # values need to be between 0 and 1
            for i in range(rank):
                for j in range(rank):
                    if W[i, j] < 0.0 or W[i, j] > 1.0:
                        W_valid_mask[i, i] -= 1

            # if the sum of the mask values equals the square of the rank,
            # no rules have been broken:
            W_valid = (np.sum(W_valid_mask) == rank ** 2)

            return W_valid, W_valid_mask

        def _validate_W(W, R):
            """Validation rules for a diagonal W matrix"""
            W_valid_mask = np.ones_like(W)
            rank = self.G ** max(R, 1)

            # sum of the diagonal nees to be one
            if abs(np.sum(W) - 1.0) > 1e6:
                for i in range(rank):
                    W_valid_mask[i, i] -= 1

            # values need to be between 0 and 1
            for i in range(rank):
                for j in range(rank):
                    if W[i, j] < 0.0 or W[i, j] > 1.0:
                        W_valid_mask[i, i] -= 1

            # if the sum of the mask values equals the square of the rank,
            # no rules have been broken:
            W_valid = (np.sum(W_valid_mask) == rank ** 2)

            return W_valid, W_valid_mask

        def _validate_P(P, R):
            P_valid_mask = np.ones_like(P)
            rank = self.G ** max(R, 1)

            # sum of the rows need to be one
            for i in range(rank):
                if abs(np.sum(P[i, ...]) - 1.0) > 1e6:
                    P_valid_mask[i, ...] -= 1

            # values need to be between 0 and 1
            for i in range(rank):
                for j in range(rank):
                    if P[i, j] < 0.0 or P[i, j] > 1.0:
                        P_valid_mask[i, j] -= 1

            # if the sum of the mask values equals the square of the rank,
            # no rules have been broken:
            P_valid = (np.sum(P_valid_mask) == rank ** 2)

            return P_valid, P_valid_mask

        for i in range(max(self.R, 1)):
            self.W_valid[i], self.W_valid_mask[i] = _validate_W(self._lW[i], i + 1)
            self.P_valid[i], self.P_valid_mask[i] = _validate_P(self._lP[i], i + 1)

        # the extra W matrix validates differently:
        self.W_valid[-1], self.W_valid_mask[-1] = _validate_WW(self._lW[-1], self.R)

    def _get_Pxy_from_indeces(self, indeces):
        if not hasattr(indeces, "__iter__"):
            indeces = [indeces]
        l = len(indeces)
        assert(l > 1), "Two or more indeces needed to acces P elements, not %s" % indeces
        assert(l <= max(self.R, 1) + 1), "Too many indeces for an R%d model: %s" % (self.R, indeces)
        R = max(l - 1, 1)
        x, y = 0, 0
        for i in range(1, R + 1):
            f = self.G ** (R - i)
            x += indeces[i - 1] * f
            y += indeces[i] * f
        return (l - 2), (x, y)

    def _get_Wxy_from_indeces(self, indeces):
        if not hasattr(indeces, "__iter__"):
            indeces = [indeces]
        l = len(indeces)
        assert(l > 0), "One or more indeces needed to acces W elements"
        assert(l <= max(self.R, 1) + 1), "Too many indeces for an R%d model: %s" % (self.R, indeces)
        if l == (max(self.R, 1) + 1):
            R = max(l - 1, 1)
            x, y = 0, 0
            for i in range(1, R + 1):
                f = self.G ** (R - i)
                x += indeces[i - 1] * f
                y += indeces[i] * f
            return (l - 1), (x, y)
        else:
            R = max(l, 1)
            x = 0
            for i in range(R):
                x += indeces[i] * self.G ** (R - (i + 1))
            return (l - 1), (x, x)

    def get_all_matrices(self): return self._lW, self._lP

    def get_distribution_matrix(self): return self._W

    def get_distribution_array(self): return np.diag(self._W)

    def get_probability_matrix(self): return self._P

    _stashed_lP = None
    _stashed_lW = None
    _stashed_flags = None
    def _stash_matrices(self):
        """ Stashes the matrices for an update """
        self._stashed_lW = deepcopy(np.asanyarray(self._lW))
        self._stashed_lP = deepcopy(np.asanyarray(self._lP))
        self._stashed_flags = deepcopy(np.asarray(self._flags))
        
    def _compare_stashed_matrices(self):
        """ Unstashed matrices and compares with current values, if identical returns True """
        if self._stashed_lP is not None and self._stashed_lW is not None:
            result = np.array_equal(self._stashed_lW, np.asanyarray(self._lW))
            result = result and np.array_equal(self._stashed_lP, np.asanyarray(self._lP))
            result = result and np.array_equal(self._stashed_flags, np.asanyarray(self._flags))
            self._stashed_lW = None
            self._stashed_lP = None
            return result
        else:
            return False
        
    @contextmanager
    def monitor_changes(self):
        with self.data_changed.hold():
            self._stash_matrices()
            yield
            if not self._compare_stashed_matrices():
                self.data_changed.emit()

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#  Copyright (C) 2005 by Roberto Cavada <roboogle@gmail.com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

from distutils.version import LooseVersion

def _cmp (self, other):
    if isinstance(other, str):
        other = LooseVersion(other)

    stypes = map(lambda c: str if isinstance(c, str) else int, self.version)
    otypes = map(lambda c: str if isinstance(c, str) else int, other.version)
    
    for i, (stype, otype) in enumerate(zip(stypes, otypes)):
        if stype == str and otype == int:
            other.version[i] = str(other.version[i])
        if stype == int and otype == str:
            self.version[i] = str(self.version[i])
    
    if self.version == other.version:
        return 0
    if self.version < other.version:
        return -1
    if self.version > other.version:
        return 1
            
LooseVersion._cmp = _cmp


# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk as gdk

import logging
logger = logging.getLogger(__name__)

from pyxrd.generic.views.combobox_tools import add_combo_text_column

from pyxrd.generic.controllers import DialogController

class AddInSituBehaviourController(DialogController):
    """ 
        Controller for the add InSituBehaviour dialog
    """

    auto_adapt = False

    def __init__(self, model=None, view=None, parent=None, callback=None):
        super(AddInSituBehaviourController, self).__init__(
            model=model, view=view, parent=parent)
        self.callback = callback

    def generate_combo(self):
        cmb_model = Gtk.ListStore(str, object)
        print("Adding rows from:", self.parent.obj_type_map)
        for cls, _, _ in self.parent.obj_type_map:
            print("Adding row:", cls)
            cmb_model.append([cls.Meta.store_id, cls])
        self.view.behaviour_combo_box.set_model(cmb_model)
        add_combo_text_column(
            self.view.behaviour_combo_box, text_col=0)

    def register_view(self, view):
        self.generate_combo()

    # ------------------------------------------------------------
    #      GTK Signal handlers
    # ------------------------------------------------------------
    def on_btn_ok_clicked(self, event):
        self.view.hide()
        self.callback(self.view.get_behaviour_type())
        return True

    def on_keypress(self, widget, event):
        if event.keyval == Gdk.KEY_Escape:
            self.view.hide()
            return True
        if event.keyval == Gdk.KEY_Return:
            return self.on_btn_ok_clicked(event)

    def on_window_edit_dialog_delete_event(self, event, args=None):
        self.view.hide()
        return True # do not propagate

    pass # end of class


# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import gi
gi.require_version('Gtk', '3.0')  # @UndefinedVariable
from gi.repository import Gtk

import logging
logger = logging.getLogger(__name__)

from mvc.models.properties import (
    StringProperty, BoolProperty, LabeledProperty, FloatProperty,
    SignalMixin, SetActionMixin
)

from pyxrd.generic.io import storables, Storable
from pyxrd.generic.models import DataModel

from pyxrd.refinement.refinables.properties import RefinableMixin
from pyxrd.refinement.refinables.mixins import RefinementValue
from pyxrd.refinement.refinables.metaclasses import PyXRDRefinableMeta

from .atom_relations import ComponentPropMixin

@storables.register()
class UnitCellProperty(ComponentPropMixin, RefinementValue, DataModel, Storable, metaclass=PyXRDRefinableMeta):
    """
        UnitCellProperty's are an integral part of a component and allow to 
        calculate the dimensions of the unit cell based on compositional
        information such as the iron content.
        This class is not responsible for keeping its value up-to-date.
        With other words, it is the responsibility of the higher-level class
        to call the 'update_value' method on this object whenever it emits a
        'data_changed' signal. The reason for this is to prevent infinite 
        recursion errors. 
    """
    class Meta(DataModel.Meta):
        store_id = "UnitCellProperty"

    component = property(DataModel.parent.fget, DataModel.parent.fset)

    # PROPERTIES:

    #: The UnitCellProperty name
    name = StringProperty(
        default="", text="Name",
        visible=False, persistent=False,
        signal_name="visuals_changed",
        mix_with=(SignalMixin,)
    )

    #: Flag indicating if this UnitCellProperty is enabled
    enabled = BoolProperty(
        default=False, text="Enabled",
        visible=True, persistent=True,
        set_action_name="update_value",
        mix_with=(SetActionMixin,)
    )

    #: Flag indicating if this UnitCellProperty is inherited
    inherited = BoolProperty(
        default=False, text="Inherited",
        visible=False, persistent=False,
        set_action_name="update_value",
        mix_with=(SetActionMixin,)
    )

    #: The value of the UnitCellProperty
    value = FloatProperty(
        default=0.0, text="Value",
        visible=True, persistent=True, refinable=True, widget_type='float_entry',
        set_action_name="update_value",
        mix_with=(SetActionMixin, RefinableMixin)
    )

    #: The factor of the UnitCellProperty (if enabled and not constant)
    factor = FloatProperty(
        default=1.0, text="Factor",
        visible=True, persistent=True, widget_type='float_entry',
        set_action_name="update_value",
        mix_with=(SetActionMixin,)
    )

    #: The constant of the UnitCellProperty (if enabled and not constant)
    constant = FloatProperty(
        default=0.0, text="Constant",
        visible=True, persistent=True, widget_type='float_entry',
        set_action_name="update_value",
        mix_with=(SetActionMixin,)
    )

    _temp_prop = None # temporary, JSON-style prop
    prop = LabeledProperty(
        default=None, text="Property",
        visible=True, persistent=True, widget_type='combo',
        set_action_name="update_value",
        mix_with=(SetActionMixin,)
    )

    # REFINEMENT VALUE IMPLEMENTATION:
    @property
    def refine_title(self):
        return self.name

    @property
    def refine_descriptor_data(self):
        return dict(
            phase_name=self.component.phase.refine_title,
            component_name=self.component.refine_title
        )

    @property
    def refine_value(self):
        return self.value
    @refine_value.setter
    def refine_value(self, value):
        if not self.enabled:
            self.value = value

    @property
    def refine_info(self):
        return self.value_ref_info

    @property
    def is_refinable(self):
        return not (self.enabled or self.inherited)

    # ------------------------------------------------------------
    #      Initialisation and other internals
    # ------------------------------------------------------------
    def __init__(self, *args, **kwargs):
        keys = [prop.label for prop in UnitCellProperty.Meta.get_local_persistent_properties()]
        keys.extend(["data_%s" % prop.label for prop in UnitCellProperty.Meta.get_local_persistent_properties()])
        my_kwargs = self.pop_kwargs(kwargs, "name", *keys)
        super(UnitCellProperty, self).__init__(*args, **kwargs)
        kwargs = my_kwargs

        with self.data_changed.hold_and_emit():
            self.name = self.get_kwarg(kwargs, self.name, "name", "data_name")
            self.value = self.get_kwarg(kwargs, self.value, "value", "data_value")
            self.factor = self.get_kwarg(kwargs, self.factor, "factor", "data_factor")
            self.constant = self.get_kwarg(kwargs, self.constant, "constant", "data_constant")
            self.enabled = self.get_kwarg(kwargs, self.enabled, "enabled", "data_enabled")

            self._temp_prop = self.get_kwarg(kwargs, self.prop, "prop", "data_prop")

    # ------------------------------------------------------------
    #      Input/Output stuff
    # ------------------------------------------------------------
    def json_properties(self):
        retval = Storable.json_properties(self)
        if retval["prop"]:
            # Try to replace objects with their uuid's:
            try:
                retval["prop"] = [getattr(retval["prop"][0], 'uuid', retval["prop"][0]), retval["prop"][1]]
            except:
                logger.exception("Error when trying to interpret UCP JSON properties")
                pass # ignore
        return retval

    def resolve_json_references(self):
        if getattr(self, "_temp_prop", None):
            self._temp_prop = list(self._temp_prop)
            if isinstance(self._temp_prop[0], str):
                obj = type(type(self)).object_pool.get_object(self._temp_prop[0])
                if obj:
                    self._temp_prop[0] = obj
                    self.prop = self._temp_prop
                else:
                    self._temp_prop = None        
            self.prop = self._temp_prop
            del self._temp_prop

    # ------------------------------------------------------------
    #      Methods & Functions
    # ------------------------------------------------------------
    def create_prop_store(self, extra_props=[]):
        assert(self.component is not None)
        store = Gtk.ListStore(object, str, str)
        # use private properties so we connect to the actual object stores and not the inherited ones
        for atom in self.component._layer_atoms:
            store.append([atom, "pn", atom.name])
        for atom in self.component._interlayer_atoms:
            store.append([atom, "pn", atom.name])
        for prop in extra_props:
            store.append(prop)
        return store

    def get_value_of_prop(self):
        try:
            return getattr(*self.prop)
        except:
            return 0.0

    def update_value(self):
        if self.enabled:
            self._value = float(self.factor * self.get_value_of_prop() + self.constant)
            self.data_changed.emit()

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import csv

class CSVMixin(object):
    """
        Model mixin providing CSV export and import functionality
    """

    @classmethod
    def save_as_csv(cls, filename, items):
        atl_writer = csv.writer(open(filename, 'w'), delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
        labels = [prop.label for prop in cls.Meta.get_local_persistent_properties()]
        atl_writer.writerow(labels)
        for item in items:
            prop_row = []
            for label in labels:
                prop_row.append(getattr(item, label))
            atl_writer.writerow(prop_row)

    @classmethod
    def get_from_csv(cls, filename, parent=None):
        with open(filename, 'r') as csvfile:
            atl_reader = csv.DictReader(csvfile, delimiter=',', quotechar='"')
            labels = [prop.label for prop in cls.Meta.get_local_persistent_properties()]
            for row in atl_reader:
                yield cls(parent=parent, **{
                    prop: row[prop] for prop in labels if prop in row
                })

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import logging
logger = logging.getLogger(__name__)

import re
from fractions import Fraction

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk, GdkPixbuf

try:
    gi.require_foreign("cairo")
except ImportError as orig:
    try:
        import cairocffi as cairo
    except ImportError as snd:
        logger.error("No cairo integration :(")
        raise snd from orig 

from matplotlib import rcParams
import matplotlib.mathtext as mathtext

pbmt_cache = dict() # maybe use a weak ref dict or a slowly GC-ed one?
display = Gdk.Display.get_default()  # @UndefinedVariable
screen = display.get_default_screen()
dpi = screen.get_resolution() or 96

def create_pb_from_mathtext(text, align='center', weight='heavy', color='b', style='normal'):
    """
        Create a Gdk.Pixbuf from a mathtext string
    """
    
    global pbmt_cache
    global dpi
    if not text in pbmt_cache:

        parts, fontsize = _handle_customs(text)

        pbs = []
        width = 0
        height = 0
        # heights = []

        # Temporarily set font properties:
        old_params = rcParams["font.weight"], rcParams["text.color"], rcParams["font.style"]
        rcParams["font.weight"] = weight
        rcParams["text.color"] = color
        rcParams["font.style"] = style
        
        # Create parser and load png fragments
        parser = mathtext.MathTextParser("Bitmap")
        for part in parts:
            png_loader = GdkPixbuf.PixbufLoader.new_with_type('png')  # @UndefinedVariable
            parser.to_png(png_loader, part, dpi=dpi, fontsize=fontsize)
            png_loader.close()
            pb = png_loader.get_pixbuf()
            w, h = pb.get_width(), pb.get_height()
            width = max(width, w)
            height += h
            pbs.append((pb, w, h))
        # Restore font properties
        rcParams["font.weight"], rcParams["text.color"], rcParams["font.style"] = old_params

        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, width, height)
        cr = cairo.Context(surface)

        cr.save()
        cr.set_operator(cairo.OPERATOR_CLEAR)
        cr.paint()
        cr.restore()

        cr.save()
        offsetx = 0
        offsety = 0
        for pb, w, h in pbs:
            if align == 'center':
                offsetx = int((width - w) / 2)
            if align == 'left':
                offsetx = 0
            if align == 'right':
                offsetx = int(width - w)
            Gdk.cairo_set_source_pixbuf(cr, pb, offsetx, offsety)
            cr.rectangle(offsetx, offsety, w, h)
            cr.paint()
            offsety += h
        del pbs
        cr.restore()

        pbmt_cache[text] = Gdk.pixbuf_get_from_surface(surface, 0, 0, width, height)
        
    return pbmt_cache[text]

def create_image_from_mathtext(text, align='center', weight='heavy', color='b', style='normal'):
    """
        Create a Gtk.Image widget from a mathtext string
    """
    
    image = Gtk.Image()
    image.set_from_pixbuf(create_pb_from_mathtext(text, align=align))
    return image

###############################
# Some convenience functions: #
###############################
def _handle_customs(text):
    text = text.decode('utf-8')

    if r"\larger" in text:
        fontsize = 20
    elif r"\large" in text:
        fontsize = 15
    else:
        fontsize = 10

    replacers = [
        (r"²", r"$^{2}$"),
        (r"³", r"$^{3}$"),
        (r"α", r"$\alpha$"),
        (r"β", r"$\beta$"),
        (r"γ", r"$\gamma$"),
        (r"δ", r"$\delta$"),
        (r"γ", r"$\digamma$"),
        (r"η", r"$\eta$"),
        (r"ι", r"$\iota$"),
        (r"κ", r"$\kappa$"),
        (r"λ", r"$\lambda$"),
        (r"μ", r"$\mu$"),
        (r"ω", r"$\omega$"),
        (r"φ", r"$\phi$"),
        (r"π", r"$\pi$"),
        (r"ψ", r"$\psi$"),
        (r"ρ", r"$\rho$"),
        (r"σ", r"$\sigma$"),
        (r"τ", r"$\tau$"),
        (r"θ", r"$\theta$"),
        (r"υ", r"$\upsilon$"),
        (r"ξ", r"$\xi$"),
        (r"ζ", r"$\zeta$"),
        (r"\larger", r""),
        (r"\large", r""),
        (r"\newline", r"$\newline$"),
    ]
    for val, rep in replacers:
        text = text.replace(val, rep)

    parts = text.replace("$$", "").split(r"\newline")
    while "$$" in parts: parts.remove("$$")
    return parts, fontsize

def mt_frac(val):
    val = Fraction(val).limit_denominator()
    if val.denominator > 1:
        return r"\frac{%d}{%d}" % (val.numerator, val.denominator)
    else:
        return r"%d" % val.numerator

def mt_range(lower, name, upper):
    return r"\left({ %s \leq %s \leq %s }\right)" % (mt_frac(lower), name, mt_frac(upper))

def get_plot_safe(expression):
    return r"".join(_handle_customs(expression)[0])

def get_string_safe(expression):

    replacers = [
        (r"$", r""),
        (r"\larger", r""),
        (r"\left", r""),
        (r"\right", r""),
        (r"\leq", r"≤"),
        (r"\geq", r"≥"),
        (r"\large", r""),
        (r"\newline", "\n"),
    ]
    for val, rep in replacers:
        expression = expression.replace(val, rep)

    regex_replacers = [
        (r"\\sum_\{(\S+)\}\^\{(\S+)\}", r"Σ(\1->\2)"),
        (r"(\S+)_(?:\{(\S+)\})", r"\1\2"),
        (r"(\S+)_(\S+)", r"\1\2"),
        (r"\\frac\{([^}])\}\{([^}])\}", r"\1\\\2"), # single characters
        (r"\\frac\{(.+)\}\{(.+)\}", r"(\1)\\(\2)"), # multi charachters
        (r"\(\{([^})]+)\}\)", r"(\1)")
    ]
    for regexpr, sub in regex_replacers:
        pattern = re.compile(regexpr)
        expression = pattern.sub(sub, expression)

    return expression


#!/usr/bin/env python3

# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import sys, os
base = os.path.dirname(__file__)
sys.path.insert(0, os.path.join(base, "pyxrd"))

if __name__ == "__main__":
    from test import run_all_tests
    run_all_tests()



#!/usr/bin/python

# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import time
import mock
import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

def create_object_attribute_test(object_name, attribute, value):
    """
        Helper function to create simple attribute setter/getter tests.
    """
    def test_attribute(self):
        obj = getattr(self, object_name)
        setattr(obj, attribute, value)
        self.assertEqual(getattr(obj, attribute), value)
    return test_attribute

# Stolen from Kiwi
def refresh_gui(delay=0):
    while Gtk.events_pending():
        Gtk.main_iteration_do(block=False)
    time.sleep(delay)
    
def _mocked_parse_args():
    args = mock.Mock()
    args.script.return_value = True
    args.script.filename = ""
    args.script.debug = False
    return args

def mock_settings():
    from pyxrd.data import settings
    settings._parse_args = mock.Mock(return_value=_mocked_parse_args())
    settings.initialize()
    


# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#  Copyright (C) 2006 by Roberto Cavada <roboogle@gmail.com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------


# This file contains decorators to be used (privately) by other parts
# of the framework

import types

def good_decorator(decorator):
    """This decorator makes decorators behave well wrt to decorated
    functions names, doc, etc."""
    def new_decorator(f):
        g = decorator(f)
        g.__name__ = f.__name__
        g.__doc__ = f.__doc__
        g.__dict__.update(f.__dict__)
        return g

    new_decorator.__name__ = decorator.__name__
    new_decorator.__doc__ = decorator.__doc__
    new_decorator.__dict__.update(decorator.__dict__)

    return new_decorator


def good_classmethod_decorator(decorator):
    """This decorator makes class method decorators behave well wrt
    to decorated class method names, doc, etc."""
    def new_decorator(cls, f):
        g = decorator(cls, f)
        g.__name__ = f.__name__
        g.__doc__ = f.__doc__
        g.__dict__.update(f.__dict__)
        return g

    new_decorator.__name__ = decorator.__name__
    new_decorator.__doc__ = decorator.__doc__
    new_decorator.__dict__.update(decorator.__dict__)

    return new_decorator


def good_decorator_accepting_args(decorator):
    """This decorator makes decorators behave well wrt to decorated
    functions names, doc, etc. 

    Differently from good_decorator, this accepts decorators possibly
    receiving arguments and keyword arguments.

    This decorato can be used indifferently with class methods and
    functions."""
    def new_decorator(*f, **k):
        g = decorator(*f, **k)
        if 1 == len(f) and isinstance(f[0], types.FunctionType):
            g.__name__ = f[0].__name__
            g.__doc__ = f[0].__doc__
            g.__dict__.update(f[0].__dict__)
            pass
        return g

    new_decorator.__name__ = decorator.__name__
    new_decorator.__doc__ = decorator.__doc__
    new_decorator.__dict__.update(decorator.__dict__)
    # Required for Sphinx' automodule.
    new_decorator.__module__ = decorator.__module__
    return new_decorator



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import logging
logger = logging.getLogger(__name__)

from .providers import get_provider
from .exceptions import *
from .dummy_async_server import DummyAsyncServer

class HasAsyncCalls(object):
    """
        Class providing utility functions for async calls
    """
    
    def submit_async_call(self, func):
        """ Utility that passes function calls down to the async server """
        try:
            self._async_server = get_provider().get_server()
        except (ServerNotRunningException, ServerStartTimeoutExcecption):
            logger.warning("Could not get the default provided async server, using dummy implementation")
            self._async_server = DummyAsyncServer()
        return self._async_server.submit(func)
        

    def fetch_async_result(self, result):
        """ Utility that parses the result objects returned by submit_async_call"""
        return result.get()

    pass #end of class